# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pprint
import re
import sys
import os

import unittest

if sys.version_info < (3, ):
    import unicodecsv as csv
else:
    import csv

import pytest

from hgvs.exceptions import HGVSError
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.sequencevariant
import hgvs.variantmapper
from support import CACHE


def gxp_file_reader(fn):
    rdr = csv.DictReader(open(fn, "r"), delimiter=str("\t"))
    for rec in rdr:
        if rec["id"].startswith("#"):
            continue
        yield rec


@pytest.mark.mapping
class Test_VariantMapper(unittest.TestCase):
    def setUp(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
        self.hm = hgvs.variantmapper.VariantMapper(self.hdp)
        self.hp = hgvs.parser.Parser()

    # ZCCHC3 -- one exon, + strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = "NM_033089.6";
    # ┌────────┬────────────┬─────────┬─────────────┬──────────────┬─────────────┬─────────────┬───────────┐
    # │  hgnc  │ alt_strand │ n_exons │    tx_ac    │    alt_ac    │  s_cigars   │ cds_start_i │ cds_end_i │
    # ├────────┼────────────┼─────────┼─────────────┼──────────────┼─────────────┼─────────────┼───────────┤
    # │ ZCCHC3 │          1 │       1 │ NM_033089.6 │ NC_000020.10 │ 484=3I2275= │          24 │      1236 │
    # └────────┴────────────┴─────────┴─────────────┴──────────────┴─────────────┴─────────────┴───────────┘
    def test_ZCCHC3_dbSNP(self):
        for rec in gxp_file_reader("tests/data/gcp/ZCCHC3-dbSNP.tsv"):
            self._test_gxp_mapping(rec)

    # ORAI1 -- two exons, + strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = "NM_032790.3";
    # ┌───────┬────────────┬─────────┬─────────────┬──────────────┬──────────────────┬─────────────┬───────────┐
    # │ hgnc  │ alt_strand │ n_exons │    tx_ac    │    alt_ac    │     s_cigars     │ cds_start_i │ cds_end_i │
    # ├───────┼────────────┼─────────┼─────────────┼──────────────┼──────────────────┼─────────────┼───────────┤
    # │ ORAI1 │          1 │       2 │ NM_032790.3 │ NC_000012.11 │ 319=6I177=;1000= │         193 │      1099 │
    # └───────┴────────────┴─────────┴─────────────┴──────────────┴──────────────────┴─────────────┴───────────┘
    def test_ORAI1_dbSNP(self):
        for rec in gxp_file_reader("tests/data/gcp/ORAI1-dbSNP.tsv"):
            self._test_gxp_mapping(rec)

    # FOLR3 -- multiple exons, + strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = "NM_000804.2";
    # ┌───────┬────────────┬─────────┬─────────────┬─────────────┬──────────────────────────────┬─────────────┬───────────┐
    # │ hgnc  │ alt_strand │ n_exons │    tx_ac    │   alt_ac    │           s_cigars           │ cds_start_i │ cds_end_i │
    # ├───────┼────────────┼─────────┼─────────────┼─────────────┼──────────────────────────────┼─────────────┼───────────┤
    # │ FOLR3 │          1 │       5 │ NM_000804.2 │ NC_000011.9 │ 44=;174=;150=2D37=;136=;304= │          50 │       788 │
    # └───────┴────────────┴─────────┴─────────────┴─────────────┴──────────────────────────────┴─────────────┴───────────┘
    def test_FOLR3_dbSNP(self):
        # TODO: CORE-158: g-to-c mapped insertions have incorrect interval bounds
        for rec in gxp_file_reader("tests/data/gcp/FOLR3-dbSNP.tsv"):
            self._test_gxp_mapping(rec)

    # ADRA2B -- one exon, - strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = "NM_000682.5";
    # ┌────────┬────────────┬─────────┬─────────────┬──────────────┬─────────────┬─────────────┬───────────┐
    # │  hgnc  │ alt_strand │ n_exons │    tx_ac    │    alt_ac    │  s_cigars   │ cds_start_i │ cds_end_i │
    # ├────────┼────────────┼─────────┼─────────────┼──────────────┼─────────────┼─────────────┼───────────┤
    # │ ADRA2B │         -1 │       1 │ NM_000682.5 │ NC_000002.11 │ 891=9D2375= │           0 │      1353 │
    # └────────┴────────────┴─────────┴─────────────┴──────────────┴─────────────┴─────────────┴───────────┘
    def test_ADRA2B_dbSNP(self):
        for rec in gxp_file_reader("tests/data/gcp/ADRA2B-dbSNP.tsv"):
            self._test_gxp_mapping(rec)

    # JRK -- multiple exons, - strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = "NM_001077527.1";
    # ┌──────┬────────────┬─────────┬────────────────┬──────────────┬───────────────────────┬─────────────┬───────────┐
    # │ hgnc │ alt_strand │ n_exons │     tx_ac      │    alt_ac    │       s_cigars        │ cds_start_i │ cds_end_i │
    # ├──────┼────────────┼─────────┼────────────────┼──────────────┼───────────────────────┼─────────────┼───────────┤
    # │ JRK  │         -1 │       3 │ NM_001077527.1 │ NC_000008.10 │ 52=;1844=2I199=;1483= │         514 │      2185 │
    # └──────┴────────────┴─────────┴────────────────┴──────────────┴───────────────────────┴─────────────┴───────────┘
    def test_JRK_dbSNP(self):
        # TODO: CORE-157: del26 on -1 strands gets reverse complemented as del62
        for rec in gxp_file_reader("tests/data/gcp/JRK-dbSNP.tsv"):
            self._test_gxp_mapping(rec)

    def test_NEFL_dbSNP(self):
        for rec in gxp_file_reader("tests/data/gcp/NEFL-dbSNP.tsv"):
            self._test_gxp_mapping(rec)

    def test_DNAH11_hgmd(self):
        for rec in gxp_file_reader("tests/data/gcp/DNAH11-HGMD.tsv"):
            self._test_gxp_mapping(rec)

    def test_DNAH11_dbSNP_NM_003777(self):
        for rec in gxp_file_reader("tests/data/gcp/DNAH11-dbSNP-NM_003777.tsv"):
            self._test_gxp_mapping(rec)

    def test_DNAH11_dbSNP_NM_001277115(self):
        for rec in gxp_file_reader("tests/data/gcp/DNAH11-dbSNP-NM_001277115.tsv"):
            self._test_gxp_mapping(rec)

    @pytest.mark.regression
    def test_regression(self):
        for rec in gxp_file_reader("tests/data/gcp/regression.tsv"):
            self._test_gxp_mapping(rec)

    @pytest.mark.extra
    def test_DNAH11_dbSNP_full(self):
        for rec in gxp_file_reader("tests/data/gcp/DNAH11-dbSNP.tsv"):
            self._test_gxp_mapping(rec)

    def test_real(self):
        for rec in gxp_file_reader("tests/data/gcp/real.tsv"):
            self._test_gxp_mapping(rec)

    def test_noncoding(self):
        for rec in gxp_file_reader("tests/data/gcp/noncoding.tsv"):
            self._test_gxp_mapping(rec)

    def _test_gxp_mapping(self, rec):
        """given one record (row) of g, c/n/r, and p (optional) test variants, map
        g->c/n/r, c/n/r->g, and c->p and verify equivalence

        """

        def _rm_del_seq(vs):
            return re.sub(r"del\w+ins", "delins", vs)

        var_g = self.hp.parse_hgvs_variant(rec["HGVSg"])
        var_x = self.hp.parse_hgvs_variant(rec["HGVSc"])
        var_p = self.hp.parse_hgvs_variant(
            rec["HGVSp"]) if rec["HGVSp"] is not None and rec["HGVSp"] != "" else None

        # g -> x
        if var_x.type == "c":
            var_x_test = self.hm.g_to_c(var_g, var_x.ac)
        elif var_x.type == "n":
            var_x_test = self.hm.g_to_n(var_g, var_x.ac)
        self.assertEqual(
            _rm_del_seq(str(var_x)),
            _rm_del_seq(str(var_x_test)),
            msg="%s != %s (g>t; %s; HGVSg=%s)" % (str(var_x_test), str(var_x), rec["id"],
                                                  rec["HGVSg"]))

        # c,n -> g
        if var_x.type == "c":
            var_g_test = self.hm.c_to_g(var_x, var_g.ac)
        elif var_x.type == "n":
            var_g_test = self.hm.n_to_g(var_x, var_g.ac)
        self.assertEqual(
            _rm_del_seq(str(var_g)),
            _rm_del_seq(str(var_g_test)),
            msg="%s != %s (t>g; %s; HGVSc=%s)" % (str(var_g_test), str(var_g), rec["id"],
                                                  rec["HGVSc"]))

        if var_p is not None:
            # c -> p
            hgvs_p_exp = str(var_p)
            var_p_test = self.hm.c_to_p(var_x, var_p.ac)

            if var_p.posedit and not var_p.posedit.uncertain:
                # if expected value isn't uncertain, strip uncertain from test
                var_p_test.posedit.uncertain = False

            hgvs_p_test = str(var_p_test)

            if re.search(r"Ter$", hgvs_p_exp):
                # if expected value doesn't have a count, strip it from the test
                hgvs_p_test = re.sub(r"Ter\d+$", "Ter", hgvs_p_test)

            self.assertEqual(
                hgvs_p_exp, hgvs_p_test, msg="%s != %s (%s)" % (hgvs_p_exp, hgvs_p_test, rec["id"]))


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import unittest

import pytest

import hgvs.dataproviders.uta

import hgvs.location
import hgvs.parser
from hgvs.exceptions import HGVSError, HGVSDataNotAvailableError, HGVSInvalidIntervalError
from hgvs.alignmentmapper import AlignmentMapper
from hgvs.enums import Datum
from support import CACHE

import hgvs

@pytest.mark.quick
class Test_AlignmentMapper(unittest.TestCase):
    ref = "GRCh37.p10"

    @classmethod
    def setUp(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
        cls.parser = hgvs.parser.Parser()

    def test_alignmentmapper_failures(self):
        hgvs.global_config.mapping.strict_bounds = True
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(self.hdp, tx_ac="bogus", alt_ac="NM_033089.6", alt_aln_method="splign")
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(
                self.hdp, tx_ac="bogus", alt_ac="NM_033089.6", alt_aln_method="transcript")
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(self.hdp, tx_ac="NM_033089.6", alt_ac="bogus", alt_aln_method="splign")
        with self.assertRaises(HGVSDataNotAvailableError):
            AlignmentMapper(
                self.hdp, tx_ac="NM_000051.3", alt_ac="NC_000011.9", alt_aln_method="bogus")
        with self.assertRaises(HGVSInvalidIntervalError):
            AlignmentMapper(self.hdp, 'NM_000348.3', 'NC_000002.11', 'splign').n_to_c(
                self.parser.parse_n_interval("-1"))
        with self.assertRaises(HGVSInvalidIntervalError):
            AlignmentMapper(self.hdp, 'NM_000348.3', 'NC_000002.11', 'splign').c_to_n(
                self.parser.parse_c_interval("99999"))

    def x_test_alignmentmapper_AlignmentMapper_LCE3C_uncertain(self):
        # ? is not yet supported
        """Use NM_178434.2 tests to test mapping with uncertain positions"""
        tx_ac = "NM_178434.2"
        alt_ac = "NC_000001.10"
        test_cases = [
            {
                "g": parser.parse_g_interval("(?_152573139)"),
                "n": parser.parse_n_interval("(?_2)"),
                "c": parser.parse_c_interval("(?_-69)")
            },
            {
                "g": parser.parse_g_interval("(152573138_?)"),
                "n": parser.parse_n_interval("(1_?)"),
                "c": parser.parse_c_interval("(-70_?)")
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_LCE3C(self):
        """NM_178434.2: LCE3C single exon, strand = +1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_178434.2"
        alt_ac = "NC_000001.10"
        test_cases = [
        # 5'
            {
                "g": self.parser.parse_g_interval("152573138"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-70")
            },
            {
                "g": self.parser.parse_g_interval("152573140"),
                "n": self.parser.parse_n_interval("3"),
                "c": self.parser.parse_c_interval("-68")
            },
        # cds
            {
                "g": self.parser.parse_g_interval("152573207"),
                "n": self.parser.parse_n_interval("70"),
                "c": self.parser.parse_c_interval("-1")
            },
            {
                "g": self.parser.parse_g_interval("152573208"),
                "n": self.parser.parse_n_interval("71"),
                "c": self.parser.parse_c_interval("1")
            },
        # 3'
            {
                "g": self.parser.parse_g_interval("152573492"),
                "n": self.parser.parse_n_interval("355"),
                "c": self.parser.parse_c_interval("285")
            },
            {
                "g": self.parser.parse_g_interval("152573493"),
                "n": self.parser.parse_n_interval("356"),
                "c": self.parser.parse_c_interval("*1")
            },
            {
                "g": self.parser.parse_g_interval("152573560"),
                "n": self.parser.parse_n_interval("423"),
                "c": self.parser.parse_c_interval("*68")
            },
            {
                "g": self.parser.parse_g_interval("152573562"),
                "n": self.parser.parse_n_interval("425"),
                "c": self.parser.parse_c_interval("*70")
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_HIST3H2A(self):
        """NM_033445.2: LCE3C single exon, strand = -1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_033445.2"
        alt_ac = "NC_000001.10"
        test_cases = [
        # 3'
            {
                "g": self.parser.parse_g_interval("228645560"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-42")
            },
            {
                "g": self.parser.parse_g_interval("228645558"),
                "n": self.parser.parse_n_interval("3"),
                "c": self.parser.parse_c_interval("-40")
            },
        # cds
            {
                "g": self.parser.parse_g_interval("228645519"),
                "n": self.parser.parse_n_interval("42"),
                "c": self.parser.parse_c_interval("-1")
            },
            {
                "g": self.parser.parse_g_interval("228645518"),
                "n": self.parser.parse_n_interval("43"),
                "c": self.parser.parse_c_interval("1")
            },
        # 5'
            {
                "g": self.parser.parse_g_interval("228645126"),
                "n": self.parser.parse_n_interval("435"),
                "c": self.parser.parse_c_interval("393")
            },
            {
                "g": self.parser.parse_g_interval("228645125"),
                "n": self.parser.parse_n_interval("436"),
                "c": self.parser.parse_c_interval("*1")
            },
            {
                "g": self.parser.parse_g_interval("228645124"),
                "n": self.parser.parse_n_interval("437"),
                "c": self.parser.parse_c_interval("*2")
            },
            {
                "g": self.parser.parse_g_interval("228645065"),
                "n": self.parser.parse_n_interval("496"),
                "c": self.parser.parse_c_interval("*61")
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_LCE2B(self):
        """NM_014357.4: LCE2B, two exons, strand = +1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_014357.4"
        alt_ac = "NC_000001.10"
        test_cases = [
        # 5'
            {
                "g": self.parser.parse_g_interval("152658599"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-54")
            },
            {
                "g": self.parser.parse_g_interval("152658601"),
                "n": self.parser.parse_n_interval("3"),
                "c": self.parser.parse_c_interval("-52")
            },
        # cds
            {
                "g": self.parser.parse_g_interval("152659319"),
                "n": self.parser.parse_n_interval("54"),
                "c": self.parser.parse_c_interval("-1")
            },
            {
                "g": self.parser.parse_g_interval("152659320"),
                "n": self.parser.parse_n_interval("55"),
                "c": self.parser.parse_c_interval("1")
            },
        # around end of exon 1
            {
                "g": self.parser.parse_g_interval("152658632"),
                "n": self.parser.parse_n_interval("34"),
                "c": self.parser.parse_c_interval("-21")
            },
            {
                "g": self.parser.parse_g_interval("152658633"),
                "n": self.parser.parse_n_interval("34+1"),
                "c": self.parser.parse_c_interval("-21+1")
            },
        # span
            {
                "g": self.parser.parse_g_interval("152658633_152659299"),
                "n": self.parser.parse_n_interval("34+1_35-1"),
                "c": self.parser.parse_c_interval("-21+1_-20-1")
            },
        # around beginning of exon 2
            {
                "g": self.parser.parse_g_interval("152659300"),
                "n": self.parser.parse_n_interval("35"),
                "c": self.parser.parse_c_interval("-20")
            },
            {
                "g": self.parser.parse_g_interval("152659299"),
                "n": self.parser.parse_n_interval("35-1"),
                "c": self.parser.parse_c_interval("-20-1")
            },
        # around end of exon 2
            {
                "g": self.parser.parse_g_interval("152659652"),
                "n": self.parser.parse_n_interval("387"),
                "c": self.parser.parse_c_interval("333")
            },
            {
                "g": self.parser.parse_g_interval("152659653"),
                "n": self.parser.parse_n_interval("388"),
                "c": self.parser.parse_c_interval("*1")
            },
        # span
            {
                "g": self.parser.parse_g_interval("152659651_152659654"),
                "n": self.parser.parse_n_interval("386_389"),
                "c": self.parser.parse_c_interval("332_*2")
            },
        # 3'
            {
                "g": self.parser.parse_g_interval("152659877"),
                "n": self.parser.parse_n_interval("612"),
                "c": self.parser.parse_c_interval("*225")
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)

    def test_alignmentmapper_AlignmentMapper_PTH2(self):
        """NM_178449.3: PTH2, two exons, strand = -1, all coordinate input/output are in HGVS"""
        tx_ac = "NM_178449.3"
        alt_ac = "NC_000019.9"
        test_cases = [
        # 3'
            {
                "g": self.parser.parse_g_interval("49926698"),
                "n": self.parser.parse_n_interval("1"),
                "c": self.parser.parse_c_interval("-102")
            },
        # cds
            {
                "g": self.parser.parse_g_interval("49926597"),
                "n": self.parser.parse_n_interval("102"),
                "c": self.parser.parse_c_interval("-1")
            },
            {
                "g": self.parser.parse_g_interval("49926596"),
                "n": self.parser.parse_n_interval("103"),
                "c": self.parser.parse_c_interval("1")
            },
        # around end of exon 1
            {
                "g": self.parser.parse_g_interval("49926469"),
                "n": self.parser.parse_n_interval("230"),
                "c": self.parser.parse_c_interval("128")
            },
            {
                "g": self.parser.parse_g_interval("49926468"),
                "n": self.parser.parse_n_interval("230+1"),
                "c": self.parser.parse_c_interval("128+1")
            },
        # span
            {
                "g": self.parser.parse_g_interval("49925901_49926467"),
                "n": self.parser.parse_n_interval("230+2_231-2"),
                "c": self.parser.parse_c_interval("128+2_129-2")
            },
        # around beginning of exon 2
            {
                "g": self.parser.parse_g_interval("49925900"),
                "n": self.parser.parse_n_interval("231-1"),
                "c": self.parser.parse_c_interval("129-1")
            },
            {
                "g": self.parser.parse_g_interval("49925899"),
                "n": self.parser.parse_n_interval("231"),
                "c": self.parser.parse_c_interval("129")
            },
        # around end of exon 2
            {
                "g": self.parser.parse_g_interval("49925725"),
                "n": self.parser.parse_n_interval("405"),
                "c": self.parser.parse_c_interval("303")
            },
            {
                "g": self.parser.parse_g_interval("49925724"),
                "n": self.parser.parse_n_interval("406"),
                "c": self.parser.parse_c_interval("*1")
            },
            {
                "g": self.parser.parse_g_interval("49925671"),
                "n": self.parser.parse_n_interval("459"),
                "c": self.parser.parse_c_interval("*54")
            },
        ]
        self.run_cases(tx_ac, alt_ac, test_cases)



    def run_cases(self, tx_ac, alt_ac, test_cases):
        am = AlignmentMapper(self.hdp, tx_ac, alt_ac, alt_aln_method="splign")
        for test_case in test_cases:
            assert test_case["c"] == am.g_to_c(test_case["g"]), f"{tx_ac}~{alt_ac} {test_case['c']} am.g_to_c failed"
            assert test_case["c"] == am.n_to_c(test_case["n"]), f"{tx_ac}~{alt_ac} {test_case['c']} am.n_to_c failed"
            assert test_case["g"] == am.c_to_g(test_case["c"]), f"{tx_ac}~{alt_ac} {test_case['g']} am.c_to_g failed"
            assert test_case["g"] == am.n_to_g(test_case["n"]), f"{tx_ac}~{alt_ac} {test_case['g']} am.n_to_g failed"
            assert test_case["n"] == am.c_to_n(test_case["c"]), f"{tx_ac}~{alt_ac} {test_case['n']} am.c_to_n failed"
            assert test_case["n"] == am.g_to_n(test_case["g"]), f"{tx_ac}~{alt_ac} {test_case['n']} am.g_to_n failed"





if __name__ == "__main__":
    unittest.main()


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import os
import re

import hgvs

# N.B. In sphinx parlance, version is supposed to be the undecorated
# version (e.g., 1.2 or 1.2.3).  release corresponds to __version__
# (e.g., 1.2.3.post3).
release = str(hgvs.__version__)
version = re.sub(r"\.post\d+$", "", release)

project = u'HGVS'
authors = project + ' Contributors'
copyright = u'2017, ' + authors

extlinks = {'issue': ('https://github.com/biocommons/hgvs/issues/%s', 'issue '), }

man_pages = [('index', project, project + u' Documentation', [project + u' Contributors'], 1)]

# even for local builds
html_theme = 'sphinx_rtd_theme'

#
# Boilerplate

autodoc_default_flags = ['members', 'undoc-members', 'show-inheritance']    #, 'inherited-members']
autoclass_content = 'both'
exclude_patterns = ['build', 'static', 'templates', 'themes']
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
html_favicon = 'static/favicon.ico'
html_logo = 'static/hgvs-logo.png'
html_static_path = ['static']
html_title = '{project} {release}'.format(project=project, release=release)
intersphinx_mapping = {'http://docs.python.org/': None, }
master_doc = 'index'
pygments_style = 'sphinx'
source_suffix = '.rst'
templates_path = ['templates']


# rst_epilog is appended to all rst files
# it's a good place to define global aliases
# If ends in .rst, sphinx will append it to itself :-(
rst_epilog_fn = os.path.join(os.path.dirname(__file__), 'rst_epilog')
rst_epilog = open(rst_epilog_fn).read()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""Utility class for creating an hgvsp SequenceVariant object,
given a transcript with variants applied.
Used in hgvsc to hgvsp conversion.

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import hgvs
from ..edit import (AAExt, AAFs, AARefAlt, AASub, Dup)
from ..exceptions import (HGVSError)
from ..location import (AAPosition, Interval)
from ..posedit import (PosEdit)
from six.moves import range

DBG = False


class AltSeqToHgvsp(object):
    def __init__(self, ref_data, alt_data):
        """Constructor

        :param ref_data: reference transcript record
        :type ref_data: recordtype
        :param alt_data: alt transcript record
        :type ref_data: recordtype
        """
        self._ref_data = ref_data
        self._alt_data = alt_data
        self._protein_accession = self._ref_data.protein_accession
        self._ref_seq = self._ref_data.aa_sequence
        self._alt_seq = self._alt_data.aa_sequence
        self._is_frameshift = self._alt_data.is_frameshift
        self._frameshift_start = self._alt_data.frameshift_start
        self._is_substitution = self._alt_data.is_substitution
        self._is_ambiguous = self._alt_data.is_ambiguous
        self._is_init_met = False

        if DBG:
            print("len ref seq:{} len alt seq:{}".format(len(self._ref_seq), len(self._alt_seq)))
            print("fs start:{} protein ac:{}".format(self._frameshift_start,
                                                     self._protein_accession))
            print(self._ref_seq)
            print(self._alt_seq)
            print("aa variant start: {}".format(self._alt_data.variant_start_aa))
            print(self._ref_data.transcript_sequence)
            print(self._alt_data.transcript_sequence)

    def build_hgvsp(self):
        """Compare two amino acid sequences; generate an hgvs tag from the output

        :return list of variants in sequence order
        :rtype list of dict
        """

        variants = []

        if not self._is_ambiguous and len(self._alt_seq) > 0:

            do_delins = True
            if self._ref_seq == self._alt_seq:
                # Silent p. variant
                start = self._alt_data.variant_start_aa
                if start - 1 < len(self._ref_seq):
                    deletion = self._ref_seq[start - 1]
                    insertion = deletion
                else:
                    start = ""
                    deletion = ""
                    insertion = ""
                    self._is_frameshift = False
                variants.append({"start": start, "ins": insertion, "del": deletion})
                do_delins = False
            elif self._is_substitution:
                if len(self._ref_seq) == len(self._alt_seq):
                    diff_pos = [(i, self._ref_seq[i], self._alt_seq[i])
                                for i in range(len(self._ref_seq))
                                if self._ref_seq[i] != self._alt_seq[i]]
                    if len(diff_pos) == 1:
                        (start, deletion, insertion) = diff_pos[0]
                        variants.append({"start": start + 1, "ins": insertion, "del": deletion})
                        do_delins = False

            if do_delins:
                if self._alt_data.is_frameshift:
                    start = self._alt_data.variant_start_aa - 1
                    while self._ref_seq[start] == self._alt_seq[start]:
                        start += 1
                    insertion = list(self._alt_seq[start:])
                    deletion = list(self._ref_seq[start:])
                    variants.append({"start": start + 1, "ins": insertion, "del": deletion})

                else:    # non-frameshifting delins or dup
                    # get size diff from diff in ref/alt lengths
                    start = self._alt_data.variant_start_aa - 1
                    delta = len(self._alt_seq) - len(self._ref_seq)
                    while self._ref_seq[start] == self._alt_seq[start]:
                        start += 1
                    offset = start + abs(delta)

                    if delta > 0:    # net insertion
                        insertion = list(self._alt_seq[start:offset])
                        deletion = []
                        ref_sub = self._ref_seq[start:]
                        alt_sub = self._alt_seq[offset:]
                    elif delta < 0:    # net deletion
                        insertion = []
                        deletion = list(self._ref_seq[start:offset])
                        ref_sub = self._ref_seq[offset:]
                        alt_sub = self._alt_seq[start:]
                    else:
                        insertion = []
                        deletion = []
                        ref_sub = self._ref_seq[start:]
                        alt_sub = self._alt_seq[start:]

                    # from start, get del/ins out to last difference
                    diff_indices = [i for i in range(len(ref_sub)) if ref_sub[i] != alt_sub[i]]
                    if not diff_indices and not deletion and insertion[0] == '*':
                        diff_indices.append(0)
                    if diff_indices:
                        max_diff = diff_indices[-1] + 1
                        insertion.extend(list(alt_sub[:max_diff]))
                        deletion.extend(list(ref_sub[:max_diff]))

                    variants.append({"start": start + 1, "ins": insertion, "del": deletion})

            if DBG:
                print(variants)

        if self._is_ambiguous:
            var_ps = [
                self._create_variant(
                    None,
                    None,
                    '',
                    '',
                    acc=self._protein_accession,
                    is_ambiguous=self._is_ambiguous)
            ]
        elif len(self._alt_seq) == 0:
            var_ps = [
                self._create_variant(
                    None,
                    None,
                    '',
                    '',
                    acc=self._protein_accession,
                    is_ambiguous=self._is_ambiguous,
                    is_no_protein=True)
            ]
        else:
            var_ps = [
                self._convert_to_sequence_variants(x, self._protein_accession) for x in variants
            ]

        if len(var_ps) > 1:
            raise HGVSError("Got multiple AA variants - not supported")
        return var_ps[0]

    #
    # internal methods
    #

    def _convert_to_sequence_variants(self, variant, acc):
        """Convert AA variant to an hgvs representation

        :param variant: contains start, del, and ins
        :type variant: dict
        :param acc: protein accession
        :type acc: str
        :return hgvs string
        :rtype str
        """
        start = variant['start']
        insertion = ''.join(variant['ins'])
        deletion = ''.join(variant['del'])

        # defaults
        is_dup = False    # assume not dup
        fsext_len = None    # fs or ext length
        is_sub = False
        is_ext = False

        if start == 1:    # initial methionine is modified
            aa_start = aa_end = AAPosition(base=start, aa=deletion)
            ref = ''
            alt = ''
            self._is_init_met = True
            self._is_ambiguous = True    # side-effect

        if insertion and insertion.find("*") == 0:    # stop codon at variant position
            aa_start = aa_end = AAPosition(base=start, aa=deletion[0])
            ref = ''
            alt = '*'
            is_sub = True

        elif start == len(self._ref_seq):    # extension
            if self._alt_seq[-1] == '*':
                fsext_len = len(insertion) - len(deletion)    # don't include the former stop codon
            else:
                fsext_len = '?'
            subst_at_stop_codon = insertion[0]

            aa_start = aa_end = AAPosition(base=start, aa='*')
            ref = ''
            alt = subst_at_stop_codon
            is_ext = True

        elif self._is_frameshift:    # frameshift
            aa_start = aa_end = AAPosition(base=start, aa=deletion[0])
            ref = ''

            try:
                fsext_len = str(
                    insertion.index("*") + 1)    # start w/ 1st change; ends w/ * (inclusive)
            except ValueError:
                fsext_len = "?"

            alt = insertion[0]

        else:    # no frameshift - sub/delins/dup
            if insertion == deletion:    # silent
                aa_start = aa_end = AAPosition(base=start, aa=deletion)
                ref = alt = ''

            elif len(insertion) == len(deletion) == 1:    # substitution
                aa_start = aa_end = AAPosition(base=start, aa=deletion)
                ref = ''
                alt = insertion
                is_sub = True

            elif len(deletion) > 0:    # delins OR deletion OR stop codon at variant position
                ref = deletion
                end = start + len(deletion) - 1
                if len(insertion) > 0:    # delins
                    aa_start = AAPosition(base=start, aa=deletion[0])
                    if end > start:
                        aa_end = AAPosition(base=end, aa=deletion[-1])
                    else:
                        aa_end = aa_start
                    alt = insertion

                else:    # deletion OR stop codon at variant position
                    if len(deletion) + start == len(
                            self._ref_seq):    # stop codon at variant position
                        aa_start = AAPosition(base=start, aa=deletion[0])
                        aa_end = AAPosition(base=start, aa=deletion[0])
                        ref = ''
                        alt = '*'
                        is_sub = True
                    else:    # deletion
                        aa_start = AAPosition(base=start, aa=deletion[0])
                        if end > start:
                            aa_end = AAPosition(base=end, aa=deletion[-1])
                        else:
                            aa_end = aa_start
                        alt = None

            elif len(deletion) == 0:    # insertion OR duplication OR extension

                is_dup, dup_start = self._check_if_ins_is_dup(start, insertion)

                if is_dup:    # duplication
                    dup_end = dup_start + len(insertion) - 1
                    aa_start = AAPosition(base=dup_start, aa=insertion[0])
                    aa_end = AAPosition(base=dup_end, aa=insertion[-1])
                    ref = alt = None

                else:    # insertion
                    start -= 1
                    end = start + 1

                    aa_start = AAPosition(base=start, aa=self._ref_seq[start - 1])
                    aa_end = AAPosition(base=end, aa=self._ref_seq[end - 1])
                    ref = None
                    alt = insertion

            else:    # should never get here
                raise ValueError("unexpected variant: {}".format(variant))

        var_p = self._create_variant(
            aa_start,
            aa_end,
            ref,
            alt,
            fsext_len=fsext_len,
            is_dup=is_dup,
            acc=acc,
            is_ambiguous=self._is_ambiguous,
            is_sub=is_sub,
            is_ext=is_ext,
            is_init_met=self._is_init_met
        )

        return var_p

    def _check_if_ins_is_dup(self, start, insertion):
        """Helper to identify an insertion as a duplicate

        :param start: 1-based insertion start
        :type start: int
        :param insertion: sequence
        :type insertion: str
        :return (is duplicate, variant start)
        :rtype (bool, int)
        """
        is_dup = False    # assume no
        variant_start = None

        dup_candidate_start = start - len(insertion) - 1
        dup_candidate = self._ref_seq[dup_candidate_start:dup_candidate_start + len(insertion)]
        if insertion == dup_candidate:
            is_dup = True
            variant_start = dup_candidate_start + 1

        return is_dup, variant_start

    def _create_variant(self,
                        start,
                        end,
                        ref,
                        alt,
                        fsext_len=None,
                        is_dup=False,
                        acc=None,
                        is_ambiguous=False,
                        is_sub=False,
                        is_ext=False,
                        is_no_protein=False,
                        is_init_met=False):
        """Creates a SequenceVariant object"""

        if is_init_met:
            posedit = AARefAlt(ref=ref, alt=alt, init_met=True)
        elif is_ambiguous:
            posedit = None
        else:
            interval = Interval(start=start, end=end)
            # Note - order matters
            if is_no_protein:
                edit = '0'
            elif is_sub:
                edit = AASub(ref=ref, alt=alt)
            elif is_ext:
                edit = AAExt(ref=ref, alt=alt, aaterm='*', length=fsext_len)
            elif self._is_frameshift:
                edit = AAFs(ref=ref, alt=alt, length=fsext_len)
            elif is_dup:
                edit = Dup()
            elif ref == alt == '':
                edit = AARefAlt(ref='', alt='')
            else:
                edit = AARefAlt(ref=ref, alt=alt)
            posedit = PosEdit(
                pos=interval,
                edit=edit,
                uncertain=hgvs.global_config.mapping.inferred_p_is_uncertain)
        var_p = hgvs.sequencevariant.SequenceVariant(acc, 'p', posedit)
        return var_p


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import pytest

import hgvs.easy
from hgvs.extras.babelfish import Babelfish


@pytest.fixture(scope="session", autouse=True)
def parser():
    return hgvs.easy.parser


@pytest.fixture(scope="session")
def am37():
    return hgvs.easy.am37


@pytest.fixture(scope="session")
def am38():
    return hgvs.easy.am38


@pytest.fixture(scope="session")
def hdp():
    return hgvs.easy.hdp


@pytest.fixture(scope="session")
def babelfish38(hdp):
    return Babelfish(hdp, assembly_name="GRCh38")



@pytest.fixture(scope="class")
def kitchen_sink_setup(request, hdp, parser, am37, am38):
    """
    Adds common fixtures to an instance passed via request.cls. 

    For example:

    @pytst.mark.usefixtures("kitchen_sink_setup")
    class MyTest(unittest.TestCase):
        def test_foo(self):
             self.parser.parse(...)

    The class must be subclass of unittest.TestCase

    See https://docs.pytest.org/en/5.3.5/unittest.html#mixing-pytest-fixtures-into-unittest-testcase-subclasses-using-marks
    
    """

    request.cls.hdp = hdp
    request.cls.parser = parser
    request.cls.am37 = am37
    request.cls.am38 = am38



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import hgvs
import hgvs.normalizer

from hgvs.exceptions import HGVSError, HGVSDataNotAvailableError, HGVSInvalidVariantError, HGVSUnsupportedOperationError
from hgvs.variantmapper import VariantMapper

_logger = logging.getLogger(__name__)


class AssemblyMapper(VariantMapper):
    """Provides simplified variant mapping for a single assembly and
    transcript-reference alignment method.

    AssemblyMapper inherits VariantMapper, which provides all
    projection functionality, and adds:
      * Automatic selection of genomic sequence accession
      * Transcript selection from genomic coordinates
      * Normalization after projection
      * Special handling for PAR regions

    AssemblyMapper is instantiated with an assembly name and
    alt_aln_method. These enable the following conveniences over
    VariantMapper:

    * The assembly and alignment method are used to
      automatically select an appropriate chromosomal reference
      sequence when mapping from a transcript to a genome (i.e.,
      c_to_g(...) and n_to_g(...)).

    * A new method, relevant_trancripts(g_variant), returns a list of
      transcript accessions available for the specified variant. These
      accessions are candidates mapping from genomic to trancript
      coordinates (i.e., g_to_c(...) and g_to_n(...)).

    Note: AssemblyMapper supports only chromosomal references (e.g.,
    NC_000006.11). It does not support contigs or other genomic
    sequences (e.g., NT_167249.1).

    """

    def __init__(self,
                 hdp,
                 assembly_name=hgvs.global_config.mapping.assembly,
                 alt_aln_method=hgvs.global_config.mapping.alt_aln_method,
                 normalize=hgvs.global_config.mapping.normalize,
                 prevalidation_level=hgvs.global_config.mapping.prevalidation_level,
                 in_par_assume=hgvs.global_config.mapping.in_par_assume,
                 replace_reference=hgvs.global_config.mapping.replace_reference,
                 add_gene_symbol=hgvs.global_config.mapping.add_gene_symbol,
                 *args,
                 **kwargs):
        """
        :param object hdp: instance of hgvs.dataprovider subclass
        :param bool replace_reference: replace reference (entails additional network access)

        :param str assembly_name: name of assembly ("GRCh38.p5")
        :param str alt_aln_method: genome-transcript alignment method ("splign", "blat", "genewise")
        :param bool normalize: normalize variants
        :param str prevalidation_level: None or Intrinsic or Extrinsic validation before mapping
        :param str in_par_assume: during x_to_g, assume this chromosome name if alignment is ambiguous

        :raises HGVSError subclasses: for a variety of mapping and data lookup failures
        """

        super(AssemblyMapper, self).__init__(
            hdp=hdp,
            replace_reference=replace_reference,
            prevalidation_level=prevalidation_level,
            add_gene_symbol=add_gene_symbol,
            *args,
            **kwargs)
        self.assembly_name = assembly_name
        self.alt_aln_method = alt_aln_method
        self.normalize = normalize
        self.in_par_assume = in_par_assume
        self._norm = None
        if self.normalize:
            self._norm = hgvs.normalizer.Normalizer(
                hdp, alt_aln_method=alt_aln_method, validate=False)
        self._assembly_map = {
            k: v
            for k, v in hdp.get_assembly_map(self.assembly_name).items() if k.startswith("NC_")
        }
        self._assembly_accessions = set(self._assembly_map.keys())

    def __repr__(self):
        return ("{self.__module__}.{t.__name__}(alt_aln_method={self.alt_aln_method}, "
                "assembly_name={self.assembly_name}, normalize={self.normalize}, "
                "prevalidation_level={self.prevalidation_level}, "
                "replace_reference={self.replace_reference})".format(self=self, t=type(self)))

    def g_to_c(self, var_g, tx_ac):
        var_out = super(AssemblyMapper, self).g_to_c(
            var_g, tx_ac, alt_aln_method=self.alt_aln_method)
        return self._maybe_normalize(var_out)

    def g_to_n(self, var_g, tx_ac):
        var_out = super(AssemblyMapper, self).g_to_n(
            var_g, tx_ac, alt_aln_method=self.alt_aln_method)
        return self._maybe_normalize(var_out)

    def g_to_t(self, var_g, tx_ac):
        var_out = super(AssemblyMapper, self).g_to_t(
            var_g, tx_ac, alt_aln_method=self.alt_aln_method)
        return self._maybe_normalize(var_out)

    def c_to_g(self, var_c):
        alt_ac = self._alt_ac_for_tx_ac(var_c.ac)
        var_out = super(AssemblyMapper, self).c_to_g(
            var_c, alt_ac, alt_aln_method=self.alt_aln_method)
        return self._maybe_normalize(var_out)

    def n_to_g(self, var_n):
        alt_ac = self._alt_ac_for_tx_ac(var_n.ac)
        var_out = super(AssemblyMapper, self).n_to_g(
            var_n, alt_ac, alt_aln_method=self.alt_aln_method)
        return self._maybe_normalize(var_out)

    def t_to_g(self, var_t):
        alt_ac = self._alt_ac_for_tx_ac(var_t.ac)
        var_out = super(AssemblyMapper, self).t_to_g(
            var_t, alt_ac, alt_aln_method=self.alt_aln_method)
        return self._maybe_normalize(var_out)

    def t_to_p(self, var_t):
        """Return a protein variant, or "non-coding" for non-coding variant types

        CAUTION: Unlike other x_to_y methods that always return
        SequenceVariant instances, this method returns a string when
        the variant type is ``n``.  This is intended as a convenience,
        particularly when looping over ``relevant_transcripts``,
        projecting with ``g_to_t``, then desiring a protein
        representation for coding transcripts.

        """
        if var_t.type == "n":
            return "non-coding"
        if var_t.type == "c":
            return self.c_to_p(var_t)
        raise HGVSInvalidVariantError("Expected a coding (c.) or non-coding (n.) variant; got " +
                                      str(var_t))

    def c_to_n(self, var_c):
        var_out = super(AssemblyMapper, self).c_to_n(var_c)
        return self._maybe_normalize(var_out)

    def n_to_c(self, var_n):
        var_out = super(AssemblyMapper, self).n_to_c(var_n)
        return self._maybe_normalize(var_out)

    def c_to_p(self, var_c):
        var_out = super(AssemblyMapper, self).c_to_p(var_c)
        return self._maybe_normalize(var_out)

    def relevant_transcripts(self, var_g):
        """return list of transcripts accessions (strings) for given variant,
        selected by genomic overlap"""
        tx = self.hdp.get_tx_for_region(var_g.ac, self.alt_aln_method, var_g.posedit.pos.start.base,
                                        var_g.posedit.pos.end.base)
        return [e["tx_ac"] for e in tx]

    def _alt_ac_for_tx_ac(self, tx_ac):
        """return chromosomal accession for given transcript accession (and
        the_assembly and aln_method setting used to instantiate this
        AssemblyMapper)

        """
        alt_acs = [
            e["alt_ac"] for e in self.hdp.get_tx_mapping_options(tx_ac) if
            e["alt_aln_method"] == self.alt_aln_method and e["alt_ac"] in self._assembly_accessions
        ]

        if not alt_acs:
            raise HGVSDataNotAvailableError("No alignments for {tx_ac} in {an} using {am}".format(
                tx_ac=tx_ac, an=self.assembly_name, am=self.alt_aln_method))

        # TODO: conditional is unnecessary; remove
        if len(alt_acs) > 1:
            names = set(self._assembly_map[ac] for ac in alt_acs)
            if names != set("XY"):
                alts = ", ".join(
                    ["{ac} ({n})".format(ac=ac, n=self._assembly_map[ac]) for ac in alt_acs])
                raise HGVSError("Multiple chromosomal alignments for {tx_ac} in {an}"
                                " using {am} (non-pseudoautosomal region) [{alts}]".format(
                                    tx_ac=tx_ac,
                                    an=self.assembly_name,
                                    am=self.alt_aln_method,
                                    alts=alts))

            # assume PAR
            if self.in_par_assume is None:
                raise HGVSError("Multiple chromosomal alignments for {tx_ac} in {an}"
                                " using {am} (likely pseudoautosomal region)".format(
                                    tx_ac=tx_ac, an=self.assembly_name, am=self.alt_aln_method))

            alt_acs = [ac for ac in alt_acs if self._assembly_map[ac] == self.in_par_assume]
            if len(alt_acs) != 1:
                raise HGVSError("Multiple chromosomal alignments for {tx_ac} in {an}"
                                " using {am}; in_par_assume={ipa} selected {n} of them".format(
                                    tx_ac=tx_ac,
                                    an=self.assembly_name,
                                    am=self.alt_aln_method,
                                    ipa=self.in_par_assume,
                                    n=len(alt_acs)))

        assert len(alt_acs) == 1, "Should have exactly one alignment at this point"
        return alt_acs[0]

    def _fetch_AlignmentMapper(self, tx_ac, alt_ac=None, alt_aln_method=None):
        """convenience version of VariantMapper._fetch_AlignmentMapper that
        derives alt_ac from transcript, assembly, and alt_aln_method
        used to instantiate the AssemblyMapper instance

        """

        if alt_ac is None:
            alt_ac = self._alt_ac_for_tx_ac(tx_ac)
        if alt_aln_method is None:
            alt_aln_method = self.alt_aln_method
        return super(AssemblyMapper, self)._fetch_AlignmentMapper(tx_ac, alt_ac, alt_aln_method)

    def _maybe_normalize(self, var):
        """normalize variant if requested, and ignore HGVSUnsupportedOperationError
        This is better than checking whether the variant is intronic because
        future UTAs will support LRG, which will enable checking intronic variants.
        """
        if self.normalize:
            try:
                return self._norm.normalize(var)
            except HGVSUnsupportedOperationError as e:
                _logger.warning(str(e) + "; returning unnormalized variant")
                # fall through to return unnormalized variant
        return var


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import unittest

import hgvs.parser
import hgvs.utils.altseqbuilder as altseqbuilder
from hgvs.utils.reftranscriptdata import RefTranscriptData

import support.mock_input_source as mock_input_data_source


class TestAltSeqBuilder(unittest.TestCase):

    # root sequence = ""
    fn = os.path.join(os.path.dirname(__file__), "data", "sanity_cp.tsv")
    _datasource = mock_input_data_source.MockInputSource(fn)
    _parser = hgvs.parser.Parser()

    def test_substitution_start(self):
        hgvsc = "NM_999999.1:c.1A>T"
        expected_sequence = "AAAATCAAATTGAAAGCGAAAGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_substitution_middle(self):
        hgvsc = "NM_999999.1:c.6A>T"
        expected_sequence = "AAAATCAAAATGAATGCGAAAGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_substitution_end(self):
        hgvsc = "NM_999999.1:c.30G>C"
        expected_sequence = "AAAATCAAAATGAAAGCGAAAGCGTTTCGCGCGAAATACGGG"
        self._run_comparison(hgvsc, expected_sequence)

    # TODO - build in support when system can handle variants in 5'utr region
    # def test_insertion_before_start(self):
    #     hgvsc = "NM_999999.1:c.-1_1insGGG"
    #     expected_sequence = "AAAATCAAAGGGATGAAAGCGAAAGCGTTTCGCGCGAAATAGGGG"
    #     self._run_comparison(hgvsc, expected_sequence)

    def test_insertion_start(self):
        hgvsc = "NM_999999.1:c.1_2insAAA"
        expected_sequence = "AAAATCAAAAAAATGAAAGCGAAAGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_insertion_middle(self):
        hgvsc = "NM_999999.1:c.22_23insT"
        expected_sequence = "AAAATCAAAATGAAAGCGAAAGCGTTTCGCGTCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_insertion_end(self):
        hgvsc = "NM_999999.1:c.29_30insGG"
        expected_sequence = "AAAATCAAAATGAAAGCGAAAGCGTTTCGCGCGAAATAGGGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    # TODO - build in support when system can handle variants in 3'utr region
    # def test_insertion_after_end(self):
    #     hgvsc = "NM_999999.1:c.30_*1insAA"
    #     expected_sequence = "AAAATCAAAATGAAAGCGAAAGCGTTTCGCGCGAAATAGAAGGGN"
    #     self._run_comparison(hgvsc, expected_sequence)

    def test_deletion_start(self):
        hgvsc = "NM_999999.1:c.1del"
        expected_sequence = "AAAATCAAATGAAAGCGAAAGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_deletion_middle(self):
        hgvsc = "NM_999999.1:c.2_7del"
        expected_sequence = "AAAATCAAAACGAAAGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_deletion_end(self):
        hgvsc = "NM_999999.1:c.30del"
        expected_sequence = "AAAATCAAAATGAAAGCGAAAGCGTTTCGCGCGAAATAGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_delins_start(self):
        hgvsc = "NM_999999.1:c.1delinsTTTT"
        expected_sequence = "AAAATCAAATTTTTGAAAGCGAAAGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_delins_middle(self):
        hgvsc = "NM_999999.1:c.2_3delinsAA"
        expected_sequence = "AAAATCAAAAAAAAAGCGAAAGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_delins_end(self):
        hgvsc = "NM_999999.1:c.30delinsCCCC"
        expected_sequence = "AAAATCAAAATGAAAGCGAAAGCGTTTCGCGCGAAATACCCCGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_dup(self):
        hgvsc = "NM_999999.1:c.16_24dup"
        expected_sequence = "AAAATCAAAATGAAAGCGAAAGCGTTTCGCGCGTTTCGCGCGAAATAGGGG"
        self._run_comparison(hgvsc, expected_sequence)

    def test_delete_gene(self):
        hgvsc = "NM_999999.1:c.-3_*1del"
        expected_sequence = ""
        self._run_comparison(hgvsc, expected_sequence)

    def test_sequence_with_length_that_is_not_divisible_by_3(self):
        hgvsc = "NM_999992.2:c.1del"
        expected_sequence = "AAAATCAAATGGGGTAGGCCCGGCAGCCAGCTTTATAGAGGAGGCAGTTTCGCC"
        with self.assertRaises(NotImplementedError):
            ac_p = "DUMMY"
            var = self._parser.parse_hgvs_variant(hgvsc)
            transcript_data = RefTranscriptData(hdp=self._datasource, tx_ac=var.ac, pro_ac=ac_p)

    # def test_2_substitutions(self):
    #     pass
    #
    # def test_2_indel_no_net_frameshift(self):
    #     pass
    #
    # def test_2_indel_net_frameshift(self):
    #     pass

    def _run_comparison(self, hgvsc, expected_sequence):

        ac_p = "DUMMY"
        var = self._parser.parse_hgvs_variant(hgvsc)
        transcript_data = RefTranscriptData(hdp=self._datasource, tx_ac=var.ac, pro_ac=ac_p)

        builder = altseqbuilder.AltSeqBuilder(var, transcript_data)
        insert_result = builder.build_altseq()
        actual_sequence = insert_result[0].transcript_sequence
        msg = "expected: {}\nactual  : {}".format(expected_sequence, actual_sequence)
        self.assertEqual(expected_sequence, actual_sequence, msg)


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os

import unittest

import pytest

from hgvs.exceptions import HGVSInvalidVariantError
import hgvs.dataproviders.uta
import hgvs.variantmapper
import hgvs.parser
import hgvs.validator
from support import CACHE

hdp = hgvs.dataproviders.uta.connect(mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)


class Test_HGVSValidator(unittest.TestCase):
    """Validator wrapper class tests (most testing is handled by the component classes)"""

    @classmethod
    def setUp(cls):
        cls.hp = hgvs.parser.Parser()
        cls.vr = hgvs.validator.Validator(hdp)

    def test_wrapper(self):
        """Test that validator wrapper is working"""
        self.assertTrue(self.vr.validate(self.hp.parse_hgvs_variant("NM_001005405.2:c.6C>A")))
        self.assertTrue(self.vr.validate(self.hp.parse_hgvs_variant("NP_000305.3:p.?")))
        self.assertTrue(self.vr.validate(self.hp.parse_hgvs_variant("NP_000542.1:p.Asn131Ser")))

    def test_not_strict_mode(self):
        """Test that validator is working when in not strict mode"""
        self.assertTrue(
            self.vr.validate(
                self.hp.parse_hgvs_variant("NM_000030.2:c.679_680+2delAAGT"), strict=False))
        with self.assertRaises(HGVSInvalidVariantError):
            self.vr.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.76_78delACTACAT"), strict=False)
        with self.assertRaises(HGVSInvalidVariantError):
            self.vr.validate(self.hp.parse_hgvs_variant("AC_01234.5:c.76_78insT"), strict=False)


@pytest.mark.quick
@pytest.mark.validation
class Test_HGVSIntrinsicValidator(unittest.TestCase):
    """Tests for internal validation"""

    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()
        cls.validate_int = hgvs.validator.IntrinsicValidator()

    def test_start_lte_end(self):
        """Test if start position is less <= end position"""
        self.assertTrue(
            self.validate_int.validate(self.hp.parse_hgvs_variant("NC_000007.13:g.36561662C>T")))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("NC_000007.13:g.36561662_36561663insT")))
        self.assertTrue(
            self.validate_int.validate(self.hp.parse_hgvs_variant("NM_01234.5:c.76_77insT")))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.123+54_123+55insT"), strict=False))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.123+54A>T"), strict=False))
        self.assertTrue(
            self.validate_int.validate(self.hp.parse_hgvs_variant("NC_012920.1:m.54G>C")))
        self.assertTrue(
            self.validate_int.validate(self.hp.parse_hgvs_variant("NC_012920.1:m.57_58insC")))

        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("NC_000007.13:g.36561664_36561663A>T"))

        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("NM_000277.1:c.3_1delAG"))

        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.123+56_123+55A>T"), strict=False)

    def test_ins_length_is_one(self):
        """Test if insertion length = 1"""
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("NC_000007.13:g.36561662_36561663insT")))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.76_77insT"), strict=False))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.123+54_123+55insT"), strict=False))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.123-54_123-53insT"), strict=False))

        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.76_78insTT"), strict=False)

        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.123+54_123+56insT"), strict=False)

    def test_del_length(self):
        """Test if del length agrees with position range"""
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.76_78delACT"), strict=False))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.123+54_123+55delTA"), strict=False))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.76_78del"), strict=False))
        self.assertTrue(
            self.validate_int.validate(self.hp.parse_hgvs_variant('NM_003060.3:c.-91_22del113')))

        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.76_78delACTACAT"), strict=False)
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("NM_000030.2:c.679_680+2delAAGT"))
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("NM_032487.4:c.831_*2687del2976"))

    def test_accession_type_pair(self):
        """Test if the accession and type of a variant match"""
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("NM_000030.2:g.679del"))
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("NM_000030.2:p.?"))
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("NP_000305.3:c.6del"))
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("NC_000007.13:c.679del"))
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_int.validate(self.hp.parse_hgvs_variant("AC_01234.5:c.679del"))
        self.assertTrue(
            self.validate_int.validate(
                self.hp.parse_hgvs_variant("AC_01234.5:c.679del"), strict=False))


@pytest.mark.validation
class Test_HGVSExtrinsicValidator(unittest.TestCase):
    """Tests for external validation"""

    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()
        cls.validate_ext = hgvs.validator.ExtrinsicValidator(hdp)

    def test_valid_ref(self):
        """Test variants with valid reference seqeuences."""
        hgvs_variants = [
            "NM_001005405.2:c.6C>A",
            "NM_001005405.2:c.-38T>A",
            "NM_001005405.2:c.*3C>G",
            "NM_001005405.2:c.435_440delCTGCTG",
            "NM_000187.3:c.457dupG",
            "NR_002728.3:r.55A>U",
            "NM_000495.3:c.610_628del19",
            "NM_000059.3:c.44_45insATT",
            "NC_000001.10:g.156104193G>A",
            "NP_000010.1:p.Asn158Ser",
            "NP_000042.3:p.His1082_Val1085delinsLeuHisGlnAla",
            "NP_000042.3:p.His1082ArgfsTer2",
        ]
        for var in [self.hp.parse_hgvs_variant(h) for h in hgvs_variants]:
            self.assertTrue(self.validate_ext.validate(var))

    def test_invalid_ref(self):
        """Test variants with invalid reference sequences."""
        hgvs_variants = [
            "NM_001005405.2:c.435_440delCTGCT",
            "NR_002728.3:r.55C>U",
            "NC_000001.10:g.156104193C>A",
            "NP_000010.1:p.Ser158Arg",
            "NP_000042.3:p.Arg1082_Val1085delinsLeuHisGlnAla",
            "NP_000042.3:p.His1082_Arg1085delinsLeuHisGlnAla",
            "NP_000042.3:p.Ser1082ArgfsTer2",
        ]

        for var in [self.hp.parse_hgvs_variant(h) for h in hgvs_variants]:
            with self.assertRaises(HGVSInvalidVariantError):
                self.validate_ext.validate(var)

    def test_c_within_cds_bound(self):
        """Test c. variants within and out of cds bound."""
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_ext.validate(self.hp.parse_hgvs_variant("NM_145901.2:c.343T>C"))
        with self.assertRaises(HGVSInvalidVariantError):
            self.validate_ext.validate(self.hp.parse_hgvs_variant("NM_145901.2:c.325C>A"))
        self.assertTrue(
            self.validate_ext.validate(self.hp.parse_hgvs_variant("NM_145901.2:c.324A>G")))
        self.assertTrue(
            self.validate_ext.validate(self.hp.parse_hgvs_variant("NM_145901.2:c.*1C>T")))
        self.assertTrue(
            self.validate_ext.validate(self.hp.parse_hgvs_variant("NM_145901.2:c.-22C>T")))


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""implements an hgvs data provider interface using UTA
(https://github.com/biocommons/uta)

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import contextlib
import inspect
import logging
import os
import re
import weakref

import psycopg2
import psycopg2.extras
import psycopg2.pool

from bioutils.assemblies import make_ac_name_map
from bioutils.digests import seq_md5
from six.moves.urllib import parse as urlparse

import hgvs
from ..dataproviders.interface import Interface
from ..exceptions import HGVSError, HGVSDataNotAvailableError
from .seqfetcher import SeqFetcher
import six

_logger = logging.getLogger(__name__)


def _stage_from_version(version):
    """return "prd", "stg", or "dev" for the given version string.  A value is always returned"""
    if version:
        m = re.match(r"^(?P<xyz>\d+\.\d+\.\d+)(?P<extra>.*)", version)
        if m:
            return "stg" if m.group("extra") else "prd"
    return "dev"


def _get_uta_db_url():
    """returns UTA DB URL based on environment variables and code version

    * if UTA_DB_URL is set, use that
    * Otherwise, if _UTA_URL_KEY is set, use that as the name of a
      config file entry and use the corresponding URL
    * Otherwise, 

    """

    if "UTA_DB_URL" in os.environ:
        return os.environ["UTA_DB_URL"]

    if "_UTA_URL_KEY" in os.environ:
        url_key = os.environ["_UTA_URL_KEY"]
    else:
        sdlc = _stage_from_version(hgvs.__version__)
        url_key = "public_{sdlc}".format(sdlc=sdlc)
    return hgvs.global_config['uta'][url_key]


def connect(db_url=None,
            pooling=hgvs.global_config.uta.pooling,
            application_name=None,
            mode=None,
            cache=None):
    """Connect to a UTA database instance and return a UTA interface instance.

    :param db_url: URL for database connection
    :type db_url: string
    :param pooling: whether to use connection pooling (postgresql only)
    :type pooling: bool
    :param application_name: log application name in connection (useful for debugging; PostgreSQL only)
    :type application_name: str

    When called with an explicit db_url argument, that db_url is used for connecting.

    When called without an explicit argument, the function default is
    determined by the environment variable UTA_DB_URL if it exists, or
    hgvs.datainterface.uta.public_db_url otherwise.

    >>> hdp = connect()
    >>> hdp.schema_version()
    '1.1'

    The format of the db_url is driver://user:pass@host/database/schema (the same
    as that used by SQLAlchemy).  Examples:

    A remote public postgresql database:
        postgresql://anonymous:anonymous@uta.biocommons.org/uta/uta_20170707'

    A local postgresql database:
        postgresql://localhost/uta_dev/uta_20170707

    For postgresql db_urls, pooling=True causes connect to use a
    psycopg2.pool.ThreadedConnectionPool.
    """

    _logger.debug('connecting to ' + str(db_url) + '...')

    if db_url is None:
        db_url = _get_uta_db_url()

    url = _parse_url(db_url)
    if url.scheme == 'sqlite':
        conn = UTA_sqlite(url, mode, cache)
    elif url.scheme == 'postgresql':
        conn = UTA_postgresql(
            url=url, pooling=pooling, application_name=application_name, mode=mode, cache=cache)
    else:
        # fell through connection scheme cases
        raise RuntimeError("{url.scheme} in {url} is not currently supported".format(url=url))
    _logger.info('connected to ' + str(db_url) + '...')
    return conn


class UTABase(Interface):
    required_version = "1.1"

    _queries = {
        "acs_for_protein_md5":
        """
            select ac
            from seq_anno
            where seq_id=?
            """,
        "gene_info":
        """
            select *
            from gene
            where hgnc=?
            """,

    # TODO: reconcile tx_exons query and build_tx_cigar
    # built_tx_cigar says it expects exons in transcript order,
    # but tx_exons isn't do that (on the - strand).
        "tx_exons":
        """
            select *
            from tx_exon_aln_v
            where tx_ac=? and alt_ac=? and alt_aln_method=?
            order by alt_start_i
            """,
        "tx_for_gene":
        """
            select hgnc, cds_start_i, cds_end_i, tx_ac, alt_ac, alt_aln_method
            from transcript T
            join exon_set ES on T.ac=ES.tx_ac where alt_aln_method != 'transcript' and hgnc=?
            """,
        "tx_for_region":
        """
            select tx_ac,alt_ac,alt_strand,alt_aln_method,min(start_i) as start_i,max(end_i) as end_i
            from exon_set ES
            join exon E on ES.exon_set_id=E.exon_set_id 
            where alt_ac=? and alt_aln_method=?
            group by tx_ac,alt_ac,alt_strand,alt_aln_method
            having min(start_i) < ? and ? <= max(end_i)
            """,
        "tx_identity_info":
        """
            select distinct(tx_ac), alt_ac, alt_aln_method, cds_start_i, cds_end_i, lengths, hgnc
            from tx_def_summary_v
            where tx_ac=?
            """,
        "tx_info":
        """
            select hgnc, cds_start_i, cds_end_i, tx_ac, alt_ac, alt_aln_method
            from transcript T
            join exon_set ES on T.ac=ES.tx_ac
            where tx_ac=? and alt_ac=? and alt_aln_method=?
            """,
        "tx_mapping_options":
        """
            select distinct tx_ac,alt_ac,alt_aln_method
            from tx_exon_aln_v where tx_ac=? and exon_aln_id is not NULL
            """,
        "tx_seq":
        """
            select seq
            from seq S
            join seq_anno SA on S.seq_id=SA.seq_id
            where ac=?
            """,
        "tx_similar":
        """
            select *
            from tx_similarity_v
            where tx_ac1 = ?
            """,
        "tx_to_pro":
        """
            select * from associated_accessions where tx_ac = ? order by pro_ac desc
            """,
    }

    def __init__(self, url, mode=None, cache=None):
        self.url = url
        self.seqfetcher = SeqFetcher()
        if mode != 'run':
            self._connect()
        super(UTABase, self).__init__(mode, cache)

    def __str__(self):
        return (
            "{n} <data_version:{dv}; schema_version:{sv}; application_name={self.application_name};"
            " url={self.url}; sequences-from={sf}>").format(
                n=type(self).__name__,
                self=self,
                dv=self.data_version(),
                sv=self.schema_version(),
                sf=os.environ.get("HGVS_SEQREPO_DIR", "seqfetcher"))

    def _fetchone(self, sql, *args):
        with self._get_cursor() as cur:
            cur.execute(sql, *args)
            return cur.fetchone()

    def _fetchall(self, sql, *args):
        with self._get_cursor() as cur:
            cur.execute(sql, *args)
            return cur.fetchall()

    ############################################################################
    # Queries

    def data_version(self):
        return self.url.schema

    def schema_version(self):
        return self._fetchone("select * from meta where key = 'schema_version'")['value']

    def get_seq(self, ac, start_i=None, end_i=None):
        return self.seqfetcher.fetch_seq(ac, start_i, end_i)

    def get_acs_for_protein_seq(self, seq):
        """
        returns a list of protein accessions for a given sequence.  The
        list is guaranteed to contain at least one element with the
        MD5-based accession (MD5_01234abc...def56789) at the end of the
        list.
        """
        md5 = seq_md5(seq)
        return [r['ac'] for r in self._fetchall(self._queries['acs_for_protein_md5'], [md5])
                ] + ['MD5_' + md5]

    def get_gene_info(self, gene):
        """
        returns basic information about the gene.

        :param gene: HGNC gene name
        :type gene: str

        # database results
        hgnc    | ATM
        maploc  | 11q22-q23
        descr   | ataxia telangiectasia mutated
        summary | The protein encoded by this gene belongs to the PI3/PI4-kinase family. This...
        aliases | AT1,ATA,ATC,ATD,ATE,ATDC,TEL1,TELO1
        added   | 2014-02-04 21:39:32.57125

        """
        return self._fetchone(self._queries['gene_info'], [gene])

    def get_tx_exons(self, tx_ac, alt_ac, alt_aln_method):
        """
        return transcript exon info for supplied accession (tx_ac, alt_ac, alt_aln_method), or None if not found

        :param tx_ac: transcript accession with version (e.g., 'NM_000051.3')
        :type tx_ac: str

        :param alt_ac: specific genomic sequence (e.g., NC_000011.4)
        :type alt_ac: str

        :param alt_aln_method: sequence alignment method (e.g., splign, blat)
        :type alt_aln_method: str

        # tx_exons = db.get_tx_exons('NM_199425.2', 'NC_000020.10', 'splign')
        # len(tx_exons)
        3

        tx_exons have the following attributes::

            {
                'tes_exon_set_id' : 98390
                'aes_exon_set_id' : 298679
                'tx_ac'           : 'NM_199425.2'
                'alt_ac'          : 'NC_000020.10'
                'alt_strand'      : -1
                'alt_aln_method'  : 'splign'
                'ord'             : 2
                'tx_exon_id'      : 936834
                'alt_exon_id'     : 2999028
                'tx_start_i'      : 786
                'tx_end_i'        : 1196
                'alt_start_i'     : 25059178
                'alt_end_i'       : 25059588
                'cigar'           : '410='
            }

        For example:

        # tx_exons[0]['tx_ac']
        'NM_199425.2'

        """
        rows = self._fetchall(self._queries['tx_exons'], [tx_ac, alt_ac, alt_aln_method])
        if len(rows) == 0:
            raise HGVSDataNotAvailableError(
                "No tx_exons for (tx_ac={tx_ac},alt_ac={alt_ac},alt_aln_method={alt_aln_method})".
                format(tx_ac=tx_ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method))

        # TODO: Check that end == transcript sequence length (but length N/A in current hdp)
        ex0 = 0 if (rows[0]["alt_strand"] == 1) else -1
        if rows[ex0]["tx_start_i"] != 0:
            raise HGVSDataNotAvailableError(
                "Alignment is incomplete; cannot use transcript for mapping"
                "(tx_ac={tx_ac},alt_ac={alt_ac},alt_aln_method={alt_aln_method})".format(
                    tx_ac=tx_ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method))
        return rows

    def get_tx_for_gene(self, gene):
        """
        return transcript info records for supplied gene, in order of decreasing length

        :param gene: HGNC gene name
        :type gene: str
        """
        return self._fetchall(self._queries['tx_for_gene'], [gene])

    def get_tx_for_region(self, alt_ac, alt_aln_method, start_i, end_i):
        """
        return transcripts that overlap given region

        :param str alt_ac: reference sequence (e.g., NC_000007.13)
        :param str alt_aln_method: alignment method (e.g., splign)
        :param int start_i: 5' bound of region
        :param int end_i: 3' bound of region
        """
        return self._fetchall(self._queries['tx_for_region'],
                              [alt_ac, alt_aln_method, start_i, end_i])

    def get_tx_identity_info(self, tx_ac):
        """returns features associated with a single transcript.

        :param tx_ac: transcript accession with version (e.g., 'NM_199425.2')
        :type tx_ac: str

        # database output
        -[ RECORD 1 ]--+-------------
        tx_ac          | NM_199425.2
        alt_ac         | NM_199425.2
        alt_aln_method | transcript
        cds_start_i    | 283
        cds_end_i      | 1003
        lengths        | {707,79,410}
        hgnc           | VSX1

        """
        rows = self._fetchall(self._queries['tx_identity_info'], [tx_ac])
        if len(rows) == 0:
            raise HGVSDataNotAvailableError(
                "No transcript definition for (tx_ac={tx_ac})".format(tx_ac=tx_ac))
        return rows[0]

    def get_tx_info(self, tx_ac, alt_ac, alt_aln_method):
        """return a single transcript info for supplied accession (tx_ac, alt_ac, alt_aln_method), or None if not found

        :param tx_ac: transcript accession with version (e.g., 'NM_000051.3')
        :type tx_ac: str

        :param alt_ac: specific genomic sequence (e.g., NC_000011.4)
        :type alt_ac: str

        :param alt_aln_method: sequence alignment method (e.g., splign, blat)
        :type alt_aln_method: str

        # database output
        -[ RECORD 1 ]--+------------
        hgnc           | ATM
        cds_start_i    | 385
        cds_end_i      | 9556
        tx_ac          | NM_000051.3
        alt_ac         | AC_000143.1
        alt_aln_method | splign

        """
        rows = self._fetchall(self._queries['tx_info'], [tx_ac, alt_ac, alt_aln_method])
        if len(rows) == 0:
            raise HGVSDataNotAvailableError(
                "No tx_info for (tx_ac={tx_ac},alt_ac={alt_ac},alt_aln_method={alt_aln_method})".
                format(tx_ac=tx_ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method))
        elif len(rows) == 1:
            return rows[0]
        else:
            raise HGVSError("Multiple ({n}) replies for tx_info(tx_ac="
                            "{tx_ac},alt_ac={alt_ac},alt_aln_method={alt_aln_method})".format(
                                n=len(rows),
                                tx_ac=tx_ac,
                                alt_ac=alt_ac,
                                alt_aln_method=alt_aln_method))

    def get_tx_mapping_options(self, tx_ac):
        """Return all transcript alignment sets for a given transcript
        accession (tx_ac); returns empty list if transcript does not
        exist.  Use this method to discovery possible mapping options
        supported in the database

        :param tx_ac: transcript accession with version (e.g., 'NM_000051.3')
        :type tx_ac: str

        # database output
        -[ RECORD 1 ]--+------------
        hgnc           | ATM
        cds_start_i    | 385
        cds_end_i      | 9556
        tx_ac          | NM_000051.3
        alt_ac         | AC_000143.1
        alt_aln_method | splign
        -[ RECORD 2 ]--+------------
        hgnc           | ATM
        cds_start_i    | 385
        cds_end_i      | 9556
        tx_ac          | NM_000051.3
        alt_ac         | NC_000011.9
        alt_aln_method | blat

        """
        rows = self._fetchall(self._queries['tx_mapping_options'], [tx_ac])
        return rows

    def get_similar_transcripts(self, tx_ac):
        """Return a list of transcripts that are similar to the given
        transcript, with relevant similarity criteria.

        >> sim_tx = hdp.get_similar_transcripts('NM_001285829.1')
        >> dict(sim_tx[0])
        { 'cds_eq': False,
        'cds_es_fp_eq': False,
        'es_fp_eq': True,
        'tx_ac1': 'NM_001285829.1',
        'tx_ac2': 'ENST00000498907' }

        where:

        * cds_eq means that the CDS sequences are identical
        * es_fp_eq means that the full exon structures are identical
          (i.e., incl. UTR)
        * cds_es_fp_eq means that the cds-clipped portions of the exon
          structures are identical (i.e., ecluding. UTR)
        * Hint: "es" = "exon set", "fp" = "fingerprint", "eq" = "equal"

        "exon structure" refers to the start and end coordinates on a
        specified reference sequence. Thus, having the same exon
        structure means that the transcripts are defined on the same
        reference sequence and have the same exon spans on that
        sequence.

        """

        rows = self._fetchall(self._queries['tx_similar'], [tx_ac])
        return rows

    def get_pro_ac_for_tx_ac(self, tx_ac):
        """Return the (single) associated protein accession for a given transcript
        accession, or None if not found."""

        rows = self._fetchall(self._queries['tx_to_pro'], [tx_ac])
        try:
            return rows[0]['pro_ac']
        except IndexError:
            return None

    def get_assembly_map(self, assembly_name):
        """return a list of accessions for the specified assembly name (e.g., GRCh38.p5)

        """
        return make_ac_name_map(assembly_name)


class UTA_postgresql(UTABase):
    def __init__(self,
                 url,
                 pooling=hgvs.global_config.uta.pooling,
                 application_name=None,
                 mode=None,
                 cache=None):
        if url.schema is None:
            raise Exception("No schema name provided in {url}".format(url=url))
        self.application_name = application_name
        self.pooling = pooling
        self._conn = None
        # If we're using connection pooling, track the set of DB
        # connections we've seen; on first use we set the schema
        # search path. Use weak references to avoid keeping connection
        # objects alive unnecessarily.
        self._conns_seen = weakref.WeakSet()
        super(UTA_postgresql, self).__init__(url, mode, cache)

    def __del__(self):
        self.close()

    def close(self):
        if self.pooling:
            self._pool.closeall()
        else:
            if self._conn is not None:
                self._conn.close()

    def _connect(self):
        if self.application_name is None:
            st = inspect.stack()
            self.application_name = os.path.basename(st[-1][1])
        conn_args = dict(
            host=self.url.hostname,
            port=self.url.port,
            database=self.url.database,
            user=self.url.username,
            password=self.url.password,
            application_name=self.application_name + "/" + hgvs.__version__,
        )
        if self.pooling:
            _logger.info("Using UTA ThreadedConnectionPool")
            self._pool = psycopg2.pool.ThreadedConnectionPool(
                hgvs.global_config.uta.pool_min, hgvs.global_config.uta.pool_max, **conn_args)
        else:
            self._conn = psycopg2.connect(**conn_args)
            self._conn.autocommit = True
            with self._get_cursor() as cur:
                self._set_search_path(cur)

        self._ensure_schema_exists()

        # remap sqlite's ? placeholders to psycopg2's %s
        self._queries = {k: v.replace('?', '%s') for k, v in six.iteritems(self._queries)}

    def _ensure_schema_exists(self):
        # N.B. On AWS RDS, information_schema.schemata always returns zero rows
        r = self._fetchone("select exists(SELECT 1 FROM pg_namespace WHERE nspname = %s)",
                           [self.url.schema])
        if r[0]:
            return
        raise HGVSDataNotAvailableError("specified schema ({}) does not exist (url={})".format(
            self.url.schema, self.url))

    @contextlib.contextmanager
    def _get_cursor(self, n_retries=1):
        """Returns a context manager for obtained from a single or pooled
        connection, and sets the PostgreSQL search_path to the schema
        specified in the connection URL.

        Although *connections* are threadsafe, *cursors* are bound to
        connections and are *not* threadsafe. Do not share cursors
        across threads.

        Use this funciton like this::

            with hdp._get_cursor() as cur:
                # your code

        Do not call this function outside a contextmanager.

        """

        n_tries_rem = n_retries + 1
        while n_tries_rem > 0:
            try:

                conn = self._pool.getconn() if self.pooling else self._conn

                # autocommit=True obviates closing explicitly
                conn.autocommit = True

                cur = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
                if self.pooling:
                    # this might be a new connection, in which case we
                    # need to set the search path
                    if conn not in self._conns_seen:
                        self._set_search_path(cur)
                        self._conns_seen.add(conn)

                yield cur

                # contextmanager executes these when context exits
                cur.close()
                if self.pooling:
                    self._pool.putconn(conn)

                break

            except psycopg2.OperationalError:

                _logger.warning(
                    "Lost connection to {url}; attempting reconnect".format(url=self.url))
                if self.pooling:
                    self._pool.closeall()
                self._connect()
                _logger.warning("Reconnected to {url}".format(url=self.url))

            n_tries_rem -= 1

        else:

            # N.B. Probably never reached
            raise HGVSError("Permanently lost connection to {url} ({n} retries)".format(
                url=self.url, n=n_retries))

    def _set_search_path(self, cur):
        cur.execute("set search_path = {self.url.schema},public;".format(self=self))

class ParseResult(urlparse.ParseResult):
    """Subclass of url.ParseResult that adds database and schema methods,
    and provides stringification.

    """

    def __new__(cls, pr):
        return super(ParseResult, cls).__new__(cls, *pr)

    @property
    def database(self):
        path_elems = self.path.split("/")
        return path_elems[1] if len(path_elems) > 1 else None

    @property
    def schema(self):
        path_elems = self.path.split("/")
        return path_elems[2] if len(path_elems) > 2 else None

    def __str__(self):
        return self.geturl()


def _parse_url(db_url):
    """parse database connection urls into components

    UTA database connection URLs follow that of SQLAlchemy, except
    that a schema may be optionally specified after the database. The
    skeleton format is:

       driver://user:pass@host/database/schema

    >>> params = _parse_url("driver://user:pass@host:9876/database/schema")

    >>> params.scheme
    u'driver'

    >>> params.hostname
    u'host'

    >>> params.username
    u'user'

    >>> params.password
    u'pass'

    >>> params.database
    u'database'

    >>> params.schema
    u'schema'

    """

    return ParseResult(urlparse.urlparse(db_url))


if __name__ == "__main__":
    import doctest
    doctest.testmod()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""implements a (position,edit) tuple that represents a localized sequence change

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import attr

from hgvs.exceptions import HGVSUnsupportedOperationError
from hgvs.enums import ValidationLevel


@attr.s(slots=True, repr=False)
class PosEdit(object):
    """
    represents a **simple** variant, consisting of a single position and edit pair
    """
    pos = attr.ib(default=None)
    edit = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def format(self, conf=None):
        """Formatting the string of PosEdit
        """
        if self.pos is None:
            rv = str(self.edit.format(conf))
        else:
            rv = "{pos}{edit}".format(pos=self.pos.format(conf), edit=self.edit.format(conf))

        if self.uncertain:
            if self.edit in ["0", ""]:
                rv = rv + "?"
            else:
                rv = "(" + rv + ")"
        return rv

    __str__ = format

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__, ", ".join(
                (a.name + "=" + str(getattr(self, a.name))) for a in self.__attrs_attrs__))

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    def length_change(self, on_error_raise=True):
        """Returns the net length change for this posedit.

        The method for computing the net length change depends on the
        type of variant (dup, del, ins, etc).  The length_change
        method hides this complexity from callers.

        :param hgvs.posedit.PosEdit self: a PosEdit instance
        :param bool on_error_raise: whether to raise an exception on errors 

        :returns: A signed int for the net change in length.  Negative \
        values imply net deletions, 0 implies a balanced insertion and \
        deletion (e.g., SNV), and positive values imply a net \
        insertion.

        :raises HGVSUnsupportedOperationError: When determining the \
        length for this variant type is ill-defined or unsupported.

        There are many circumstances in which the net length change
        cannot be determined, is ill-defined, or is unsupported.  In
        these cases, the result depends on the value of
        `on_error_raise`: when `on_error_raise` is True, an exception
        is raised; when False, the exception is caught and `None` is
        returned.  Callers might wish to pass `on_error_raise=False`
        in list comprehensions to avoid dealing with exceptions.

        """

        try:
            ilen = self.pos._length()
            (del_len, ins_len) = self.edit._del_ins_lengths(ilen)
            return ins_len - del_len
        except HGVSUnsupportedOperationError:
            if on_error_raise:
                raise
            return None

    def validate(self):
        if self.pos:
            (res, msg) = self.pos.validate()
            if res != ValidationLevel.VALID:
                return (res, msg)
            try:
                # Check ins length is 1
                if self.edit.type == "ins" and self.pos.end - self.pos.start != 1:
                    return (ValidationLevel.ERROR, "insertion length must be 1")
                # Check del length
                if self.edit.type in ["del", "delins"]:
                    ref_len = self.edit.ref_n
                    if ref_len is not None and ref_len != self.pos.end - self.pos.start + 1:
                        return (ValidationLevel.ERROR,
                                "Length implied by coordinates must equal sequence deletion length")
            except HGVSUnsupportedOperationError as err:
                return (ValidationLevel.WARNING, str(err))
        return (ValidationLevel.VALID, None)


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
import re
import unittest

from hgvs.exceptions import HGVSDataNotAvailableError
import hgvs.dataproviders.uta
import hgvs.edit
import hgvs.parser
import hgvs.sequencevariant
import hgvs.variantmapper
from six.moves import map


class LineIterator(object):
    """iterate over a stream, keeping last line and # lines read"""

    def __init__(self, fh, skip_comments=False):
        self.fh = fh
        self.lines_read = None
        self.last_line = None
        self.skip_comments = skip_comments

    def __iter__(self):
        self.lines_read = 0
        for line in self.fh:
            self.lines_read += 1
            if self.skip_comments and line.startswith("#"):
                continue
            self.last_line = line
            yield line


class CrossChecker(object):
    """base class for testing a group of related variants.  Checks all reasonable combinations of
    g <-> t --> p

    """

    def crosscheck_variant_group(self, variants):
        """crosscheck a group of variants; returns None if successful, otherwise a message"""

        assert all(isinstance(v, hgvs.sequencevariant.SequenceVariant) for v in variants)

        variants = sorted(variants, key=lambda v: v.type)
        binned_variants = {k: [] for k in "cgmnrp"}
        binned_variants.update(
            {g: list(gi)
             for g, gi in itertools.groupby(variants, lambda v: v.type)})
        binned_variants["t"] = binned_variants["c"] + binned_variants["n"]

        assert len(binned_variants["g"]) == len(
            [v.ac for v in binned_variants["g"]]), "variants have multiple alignments"

        # g -> t: for each g., map to each transcript accession.
        for g_var in binned_variants["g"]:
            for t_var in binned_variants["t"]:
                try:
                    r = self.vm.g_to_t(g_var, t_var.ac)
                except HGVSDataNotAvailableError:
                    continue
                if t_var != r:
                    return "g_to_t({g_var},{t_var.ac}): got {r}; expected {t_var}".format(
                        g_var=g_var, t_var=t_var, r=r)

        # t -> g: for each t., map to each genomic accession
        for t_var in binned_variants["t"]:
            for g_var in binned_variants["g"]:
                try:
                    r = self.vm.t_to_g(t_var, g_var.ac)
                except HGVSDataNotAvailableError:
                    continue
                if g_var != r:
                    return "t_to_g({t_var},{g_var.ac}): got {r}; expected {g_var}".format(
                        g_var=g_var, t_var=t_var, r=r)

        # c -> p: for each c., map to a protein variant and check whether it's in result set

        if binned_variants["p"]:
            for c_var in binned_variants["c"]:
                try:
                    r = self.vm.c_to_p(c_var)
                except HGVSDataNotAvailableError:
                    continue
                r.posedit.uncertain = False
                if isinstance(r.posedit.edit, hgvs.edit.AAFs):
                    r.posedit.edit.length = None    # Clinvar doesn't have distance to Ter, so remove it
                if r not in binned_variants["p"]:
                    return "c_to_p({c_var}): got {r}; expected on of {p_vars}".format(
                        c_var=c_var, p_vars=" ".join(map(str, binned_variants["p"])), r=r)

        return None


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""represents simple sequence-based variants

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import attr
import re

import hgvs.variantmapper
from hgvs.enums import ValidationLevel
from hgvs.utils.validation import validate_type_ac_pair


@attr.s(slots=True, repr=False)
class SequenceVariant(object):
    """
    represents a basic HGVS variant.  The only requirement is that each
    component can be stringified; for example, passing pos as either a string
    or an hgvs.location.CDSInterval (for example) are both intended uses
    """
    ac = attr.ib()
    type = attr.ib()
    posedit = attr.ib()
    gene = attr.ib(default=None)

    def format(self, conf=None):
        """Formatting the stringification of sequence variants

        :param conf: a dict comprises formatting options. None is to use global settings.

        See :class:`hgvs.config`.
        """

        if self.ac:
            ref = self.ac
            if self.gene:
                ref += "(" + self.gene + ")"
            ref += ":"
        else:
             ref = ""   

        if self.posedit is not None:
            posedit = self.posedit.format(conf)
        else:
            posedit = "?"
        typo = "{type}.{posedit}".format(type=self.type, posedit=posedit)

        return ref + typo

    __str__ = format

    def __repr__(self):
        return "{0}({1})".format(
            self.__class__.__name__, ", ".join(
                (a.name + "=" + str(getattr(self, a.name))) for a in self.__attrs_attrs__))

    def fill_ref(self, hdp):
        # TODO: Refactor. SVs should not operate on themselves when
        # external resources are required
        hm = hgvs.variantmapper.VariantMapper(hdp)
        type = None
        if isinstance(self.posedit, hgvs.posedit.PosEdit) and isinstance(
                self.posedit.edit, hgvs.edit.Edit):
            type = self.posedit.edit.type
        if type in ["del", "delins", "identity", "dup", "repeat"
                    ] and self.posedit.edit.ref_s is None:
            hm._replace_reference(self)
        if type == "identity" and isinstance(self.posedit.edit, hgvs.edit.NARefAlt):
            self.posedit.edit.alt = self.posedit.edit.ref
        return self

    def validate(self):
        (res, msg) = (ValidationLevel.VALID, None)
        if self.ac and self.type:
            (res, msg) = validate_type_ac_pair(self.type, self.ac)
            if res == ValidationLevel.ERROR:
                return (res, msg)
        if self.posedit is None:
            return (res, msg)
        (pe_res, pe_msg) = self.posedit.validate()
        if pe_res == ValidationLevel.VALID:
            return (res, msg)
        return (pe_res, pe_msg)


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import pytest


from hgvs.exceptions import HGVSInvalidIntervalError
from hgvs.utils.cigarmapper import CIGARMapper


cigar = "3=2N=X=3N=I=D="
#  0   1   2           3   4   5               6       7   8   9  tgt
#  =   =   =   N   N   =   X   =   N   N   N   =   I   =   D   =
#  0   1   2   3   4   5   6   7   8   9  10  11  12  13      14  ref


cm = CIGARMapper(cigar)


def test_cigarmapper():
    """simple cigar test
    
    """

    assert cm.ref_len == 15
    assert cm.tgt_len == 10
    
    assert len(cm.ref_pos) == len(cm.tgt_pos)
    assert cm.ref_pos == [0, 3, 5, 6, 7, 8, 11, 12, 13, 14, 14, 15]
    assert cm.tgt_pos == [0, 3, 3, 4, 5, 6,  6,  7,  7,  8,  9, 10]

    # ref to tgt
    assert cm.map_ref_to_tgt(0,'start') == (0, 0, '=')
    assert cm.map_ref_to_tgt(0,'end') == (0, 0, '=')
    assert cm.map_ref_to_tgt(1,'start') == (1, 0, '=')
    assert cm.map_ref_to_tgt(1,'end') == (1, 0, '=')
    assert cm.map_ref_to_tgt(2,'start') == (2, 0, '=')
    assert cm.map_ref_to_tgt(2,'end') == (2, 0, '=')
    assert cm.map_ref_to_tgt(3,'start') == (2, 1, 'N')
    assert cm.map_ref_to_tgt(3,'end') == (2, 1, 'N')
    assert cm.map_ref_to_tgt(4,'start') == (3, -1, 'N')
    assert cm.map_ref_to_tgt(4,'end') == (3, -1, 'N')
    assert cm.map_ref_to_tgt(5,'start') == (3, 0, '=')
    assert cm.map_ref_to_tgt(5,'end') == (3, 0, '=')
    assert cm.map_ref_to_tgt(6,'start') == (4, 0, 'X')
    assert cm.map_ref_to_tgt(6,'end') == (4, 0, 'X')
    assert cm.map_ref_to_tgt(7,'start') == (5, 0, '=')
    assert cm.map_ref_to_tgt(7,'end') == (5, 0, '=')
    assert cm.map_ref_to_tgt(8,'start') == (5, 1, 'N')
    assert cm.map_ref_to_tgt(8,'end') == (5, 1, 'N')
    assert cm.map_ref_to_tgt(9,'start') == (5, 2, 'N')
    assert cm.map_ref_to_tgt(9,'end') == (5, 2, 'N')
    assert cm.map_ref_to_tgt(10,'start') == (6, -1, 'N')
    assert cm.map_ref_to_tgt(10,'end') == (6, -1, 'N')
    assert cm.map_ref_to_tgt(11,'start') == (6, 0, '=')
    assert cm.map_ref_to_tgt(11,'end') == (6, 0, '=')
    assert cm.map_ref_to_tgt(12,'start') == (6, 0, 'I')
    assert cm.map_ref_to_tgt(12,'end') == (7, 0, 'I')
    assert cm.map_ref_to_tgt(13,'start') == (7, 0, '=')
    assert cm.map_ref_to_tgt(13,'end') == (7, 0, '=')
    assert cm.map_ref_to_tgt(14,'start') == (9, 0, '=')
    assert cm.map_ref_to_tgt(14,'end') == (9, 0, '=')

    # tgt to ref
    assert cm.map_tgt_to_ref(0,'start') == (0, 0, '=')
    assert cm.map_tgt_to_ref(0,'end') == (0, 0, '=')
    assert cm.map_tgt_to_ref(1,'start') == (1, 0, '=')
    assert cm.map_tgt_to_ref(1,'end') == (1, 0, '=')
    assert cm.map_tgt_to_ref(2,'start') == (2, 0, '=')
    assert cm.map_tgt_to_ref(2,'end') == (2, 0, '=')
    assert cm.map_tgt_to_ref(3,'start') == (5, 0, '=')
    assert cm.map_tgt_to_ref(3,'end') == (5, 0, '=')
    assert cm.map_tgt_to_ref(4,'start') == (6, 0, 'X')
    assert cm.map_tgt_to_ref(4,'end') == (6, 0, 'X')
    assert cm.map_tgt_to_ref(5,'start') == (7, 0, '=')
    assert cm.map_tgt_to_ref(5,'end') == (7, 0, '=')
    assert cm.map_tgt_to_ref(6,'start') == (11, 0, '=')
    assert cm.map_tgt_to_ref(6,'end') == (11, 0, '=')
    assert cm.map_tgt_to_ref(7,'start') == (13, 0, '=')
    assert cm.map_tgt_to_ref(7,'end') == (13, 0, '=')
    assert cm.map_tgt_to_ref(8,'start') == (13, 0, 'D')
    assert cm.map_tgt_to_ref(8,'end') == (14, 0, 'D')
    assert cm.map_tgt_to_ref(9,'start') == (14, 0, '=')
    assert cm.map_tgt_to_ref(9,'end') == (14, 0, '=')


    
def test_cigarmapper_strict_bounds():
    # exception raised for out of bounds on left?
    with pytest.raises(HGVSInvalidIntervalError):
        cm.map_ref_to_tgt(-1, "start", strict_bounds=True)

    # ... and right?
    with pytest.raises(HGVSInvalidIntervalError):
        cm.map_ref_to_tgt(cm.ref_len + 1, "start", strict_bounds=True)

    # test whether 1 base outside bounds results in correct position
    assert ( 0, 0, '=') == cm.map_ref_to_tgt( 0, "start", strict_bounds=True)
    assert (-1, 0, '=') == cm.map_ref_to_tgt(-1, "start", strict_bounds=False)
    assert (cm.tgt_len    , 0, '=') == cm.map_ref_to_tgt(cm.ref_len,     "start", strict_bounds=True)
    assert (cm.tgt_len - 1, 0, '=') == cm.map_ref_to_tgt(cm.ref_len - 1, "start", strict_bounds=False)
    



# -*- coding: utf-8 -*-
"""Representation of edit operations in HGVS variants

NARefAlt and AARefAlt are abstractions of several major variant
types.  They are distinguished by whether the ref and alt elements
of the structure.  The HGVS grammar for NA and AA are subtly
different (e.g., the ref AA in a protein substitution is part of the
location).

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import attr

from bioutils.sequences import aa_to_aa1, aa1_to_aa3

import hgvs
from hgvs.exceptions import HGVSError, HGVSUnsupportedOperationError
import six


@attr.s(slots=True)
class Edit(object):
    def format(self, conf=None):
        return str(self)

    def _format_config_na(self, conf=None):
        max_ref_length = hgvs.global_config.formatting.max_ref_length
        if conf and "max_ref_length" in conf:
            max_ref_length = conf["max_ref_length"]
        return max_ref_length

    def _format_config_aa(self, conf=None):
        p_3_letter = hgvs.global_config.formatting.p_3_letter
        p_term_asterisk = hgvs.global_config.formatting.p_term_asterisk
        p_init_met = hgvs.global_config.formatting.p_init_met

        if conf and "p_3_letter" in conf and conf["p_3_letter"] is not None:
            p_3_letter = conf["p_3_letter"]
        if conf and "p_term_asterisk" in conf and conf["p_term_asterisk"] is not None:
            p_term_asterisk = conf["p_term_asterisk"]
        if conf and "p_init_met" in conf and conf["p_init_met"] is not None:
            p_init_met = conf["p_init_met"]
        return p_3_letter, p_term_asterisk, p_init_met

    def _del_ins_lengths(self, ilen):
        raise HGVSUnsupportedOperationError(
            "internal function _del_ins_lengths not implemented for this variant type")


@attr.s(slots=True)
class NARefAlt(Edit):
    """
    represents substitutions, deletions, insertions, and indels.

    :ivar ref: reference sequence or length
    :ivar alt: alternate sequence
    :ivar uncertain: boolean indicating whether the variant is uncertain/undetermined
    """
    ref = attr.ib(default=None)
    alt = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    @property
    def ref_s(self):
        """
        returns a string representing the ref sequence, if it is not None and smells like a sequence

        >>> NARefAlt("ACGT").ref_s
        u'ACGT'
        >>> NARefAlt("7").ref_s
        >>> NARefAlt(7).ref_s

        """
        return self.ref if (isinstance(self.ref, six.string_types) and self.ref
                            and self.ref[0] in "ACGTUN") else None

    @property
    def ref_n(self):
        """
        returns an integer, either from the `ref` instance variable if it's a number, or the length of
        ref if it's a string, or None otherwise

        >>> NARefAlt("ACGT").ref_n
        4
        >>> NARefAlt("7").ref_n
        7
        >>> NARefAlt(7).ref_n
        7

        """
        try:
            return int(self.ref)
        except ValueError:
            return len(self.ref) if self.ref else None

    def format(self, conf=None):
        if self.ref is None and self.alt is None:
            raise HGVSError("RefAlt: ref and alt sequences are both undefined")

        max_ref_length = self._format_config_na(conf)

        if max_ref_length is not None:
            ref = self.ref_s
            if ref is None or len(ref) > max_ref_length:
                ref = ''
        else:
            ref = self.ref

        # subst and delins
        if self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                s = "{ref}=".format(ref=ref)
            elif len(self.alt) == 1 and len(
                    self.ref) == 1 and not self.ref.isdigit():    # don't turn del5insT into 5>T
                s = "{self.ref}>{self.alt}".format(self=self)
            else:
                s = "del{ref}ins{alt}".format(ref=ref, alt=self.alt)
        # del case
        elif self.ref is not None:
            s = "del{ref}".format(ref=ref)

        # ins case
        else:    # self.alt is not None
            s = "ins{self.alt}".format(self=self)

        return "(" + s + ")" if self.uncertain else s

    __str__ = format

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        if self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                edit_type = "identity"
            elif len(self.alt) == 1 and len(self.ref) == 1 and not self.ref.isdigit():
                edit_type = "sub"
            else:
                edit_type = "delins"
        elif self.ref is not None:
            edit_type = "del"
        else:
            edit_type = "ins"
        return edit_type

    def _del_ins_lengths(self, ilen):
        """returns (del_len, ins_len).
        Unspecified ref or alt returns None for del_len or ins_len respectively.
        """
        if self.ref == self.alt:
            del_len = ins_len = 0
        else:
            del_len = 0 if self.ref is None else ilen
            ins_len = 0 if self.alt is None else len(self.alt)
        return (del_len, ins_len)


@attr.s(slots=True)
class AARefAlt(Edit):
    ref = attr.ib(default=None)
    alt = attr.ib(default=None)
    uncertain = attr.ib(default=False)
    init_met = attr.ib(default=False)

    def __attrs_post_init__(self):
        self.ref = aa_to_aa1(self.ref)
        self.alt = aa_to_aa1(self.alt)

    def format(self, conf=None):
        if self.ref is None and self.alt is None:
            # raise HGVSError("RefAlt: ref and alt sequences are both undefined")
            return "="

        p_3_letter, p_term_asterisk, p_init_met = self._format_config_aa(conf)

        if self.init_met and p_init_met:
            s = "Met1?"
        elif self.init_met and not p_init_met:
            s = "?"
        # subst and delins
        elif self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                if p_3_letter:
                    s = "{ref}=".format(ref=aa1_to_aa3(self.ref))
                    if p_term_asterisk and s == "Ter=":
                        s = "*="
                else:
                    s = "{ref}=".format(ref=self.ref)
            elif len(self.ref) == 1 and len(self.alt) == 1:
                if p_3_letter:
                    s = aa1_to_aa3(self.alt)
                    if p_term_asterisk and s == "Ter":
                        s = "*"
                else:
                    s = self.alt
            else:
                if p_3_letter:
                    s = "delins{alt}".format(alt=aa1_to_aa3(self.alt))
                    if p_term_asterisk and s == "delinsTer":
                        s = "delins*"
                else:
                    s = "delins{alt}".format(alt=self.alt)

        # del case
        elif self.ref is not None and self.alt is None:
            s = "del"

        # ins case
        elif self.ref is None and self.alt is not None:
            if p_3_letter:
                s = "ins{alt}".format(alt=aa1_to_aa3(self.alt))
                if p_term_asterisk and s == "insTer":
                    s = "ins*"
            else:
                s = "ins{alt}".format(alt=self.alt)

        else:
            raise RuntimeError("Should not be here")

        return "(" + s + ")" if self.uncertain else s

    __str__ = format

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        if self.ref is not None and self.alt is not None:
            if self.ref == self.alt:
                edit_type = "identity"
            elif len(self.ref) == 1 and len(self.alt) == 1:
                edit_type = "sub"
            else:
                edit_type = "delins"
        elif self.ref is not None and self.alt is None:
            edit_type = "del"
        elif self.ref is None and self.alt is not None:
            edit_type = "ins"
        return edit_type

    def _del_ins_lengths(self, ilen):
        """returns (del_len, ins_len).
        Unspecified ref or alt returns None for del_len or ins_len respectively.
        """
        if self.ref == self.alt:
            del_len = ins_len = 0
        else:
            del_len = 0 if (self.ref is None or self.alt == "") else ilen
            ins_len = 0 if self.alt is None else len(self.alt)
        return (del_len, ins_len)


@attr.s(slots=True)
class AASub(AARefAlt):
    def format(self, conf=None):
        p_3_letter, p_term_asterisk, p_init_met = self._format_config_aa(conf)

        if p_3_letter:
            s = aa1_to_aa3(self.alt) if self.alt != "?" else self.alt
            if p_term_asterisk and s == "Ter":
                s = "*"
        else:
            s = self.alt
        return "(" + s + ")" if self.uncertain else s

    __str__ = format

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "sub"


@attr.s(slots=True)
class AAFs(Edit):
    ref = attr.ib(default=None)
    alt = attr.ib(default=None)
    length = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def __attrs_post_init__(self):
        self.ref = aa_to_aa1(self.ref)
        self.alt = aa_to_aa1(self.alt)

    def format(self, conf=None):
        p_3_letter, p_term_asterisk, p_init_met = self._format_config_aa(conf)

        st_length = self.length or ""
        if p_3_letter:
            if p_term_asterisk:
                s = "{alt}fs*{length}".format(alt=aa1_to_aa3(self.alt), length=st_length)
            else:
                s = "{alt}fsTer{length}".format(alt=aa1_to_aa3(self.alt), length=st_length)
        else:
            s = "{alt}fs*{length}".format(alt=self.alt, length=st_length)
        return "(" + s + ")" if self.uncertain else s

    __str__ = format

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "fs"


@attr.s(slots=True)
class AAExt(Edit):
    ref = attr.ib(default=None)
    alt = attr.ib(default=None)
    aaterm = attr.ib(default=None)
    length = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def __attrs_post_init__(self):
        self.ref = aa_to_aa1(self.ref)
        self.alt = aa_to_aa1(self.alt)
        self.aaterm = aa_to_aa1(self.aaterm)

    def format(self, conf=None):
        p_3_letter, p_term_asterisk, p_init_met = self._format_config_aa(conf)

        st_alt = self.alt or ""
        st_aaterm = self.aaterm or ""
        st_length = self.length or ""
        if p_3_letter:
            st_alt = aa1_to_aa3(st_alt)
            st_aaterm = aa1_to_aa3(st_aaterm)
            if p_term_asterisk and st_alt == "Ter":
                st_alt = "*"
            if p_term_asterisk and st_aaterm == "Ter":
                st_aaterm = "*"

        s = "{alt}ext{term}{length}".format(alt=st_alt, term=st_aaterm, length=st_length)
        return "(" + s + ")" if self.uncertain else s

    __str__ = format

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "ext"

    def _del_ins_lengths(self, ilen):
        """returns (del_len, ins_len).
        Unspecified ref or alt returns None for del_len or ins_len respectively.
        """
        return (0, abs(self.length))


@attr.s(slots=True)
class Dup(Edit):
    ref = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def format(self, conf=None):
        max_ref_length = self._format_config_na(conf)
        if max_ref_length is not None:
            ref = self.ref_s
            if ref is None or len(ref) > max_ref_length:
                ref = ''
        else:
            ref = self.ref
        return "dup" + (ref or "")

    __str__ = format

    @property
    def ref_s(self):
        """
        returns a string representing the ref sequence, if it is not None and smells like a sequence
        """
        return self.ref if (isinstance(self.ref, six.string_types) and self.ref
                            and self.ref[0] in "ACGTUN") else None

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "dup"

    def _del_ins_lengths(self, ilen):
        """returns (del_len, ins_len).
        Unspecified ref or alt returns None for del_len or ins_len respectively.
        """
        if self.ref is not None and self.ref != "":
            assert len(self.ref) == ilen
        return (0, ilen)


@attr.s(slots=True)
class Repeat(Edit):
    ref = attr.ib(default=None)
    min = attr.ib(default=None)
    max = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def format(self, conf=None):
        if self.min > self.max:
            raise HGVSError("Repeat min count must be less than or equal to max count")
        max_ref_length = self._format_config_na(conf)
        ref = self.ref
        if max_ref_length is not None and (ref is None or len(ref) > max_ref_length):
            ref = ''
        if self.min == self.max:
            return "{ref}[{min}]".format(ref=ref, min=self.min)
        return "{ref}({min}_{max})".format(ref=ref, min=self.min, max=self.max)

    __str__ = format

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "repeat"


@attr.s(slots=True)
class NACopy(Edit):
    """Represent copy number variants (Invitae-specific use)

    This class is intended for Invitae use only and does not represent
    a standard HGVS concept. The class may be changed, moved, or
    removed without notice.

    """
    copy = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def __str__(self):
        s = "copy{}".format(self.copy)
        return "(" + s + ")" if self.uncertain else s

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "copy"

    def _del_ins_lengths(self, ilen):
        """returns (del_len, ins_len).
        Unspecified ref or alt returns None for del_len or ins_len respectively.
        """
        return (0, ilen * self.copy)


@attr.s(slots=True)
class Inv(Edit):
    """Inversion
    """
    ref = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def __str__(self):
        return "inv"

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def ref_s(self):
        return self.ref if (isinstance(self.ref, six.string_types) and self.ref
                            and self.ref[0] in "ACGTUN") else None

    @property
    def ref_n(self):
        """
        returns an integer, either from the `seq` instance variable if it's a number,
        or None otherwise
        """
        try:
            return int(self.ref)
        except ValueError:
            return None

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "inv"

    def _del_ins_lengths(self, ilen):
        """returns (del_len, ins_len).
        Unspecified ref or alt returns None for del_len or ins_len respectively.
        """
        return (ilen, ilen)


@attr.s(slots=True)
class Conv(Edit):
    """Conversion
    """
    from_ac = attr.ib(default=None)
    from_type = attr.ib(default=None)
    from_pos = attr.ib(default=None)
    uncertain = attr.ib(default=False)

    def __str__(self):
        if self.from_ac and self.from_type and self.from_pos:
            s = "con{self.from_ac}:{self.from_type}.{self.from_pos}".format(self=self)
        else:
            s = "con"
        return "(" + s + ")" if self.uncertain else s

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self

    @property
    def type(self):
        """return the type of this Edit

        :returns: edit type (str)
        """
        return "con"


if __name__ == "__main__":
    import doctest
    doctest.testmod()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# we're not testing hgvs.parser, but rather merely using it to load the
# grammar.  See test_hgvs_parser.py for the parser tests

import unittest

import pytest

import hgvs.parser


@pytest.mark.quick
class Test_Parser(unittest.TestCase):
    longMessage = True

    def setUp(self):
        self.p = hgvs.parser.Parser(expose_all_rules=True)
        self.grammar = self.p._grammar

    def test_parser_basic(self):
        # num & snum
        self.assertEqual(self.grammar("55").num(), 55)
        self.assertEqual(self.grammar("55").snum(), 55)
        self.assertEqual(self.grammar("-55").snum(), -55)

        # accn
        self.assertEqual(self.grammar("NM_01234.5").accn(), "NM_01234.5")
        self.assertEqual(self.grammar("NM_01234").accn(), "NM_01234")
        self.assertEqual(self.grammar("ENST01234").accn(), "ENST01234")
        self.assertEqual(self.grammar("LRG_1234").accn(), "LRG_1234")
        self.assertEqual(self.grammar("LRG_1234t1").accn(), "LRG_1234t1")

    def test_parser_aa(self):
        # terms
        self.assertEqual(self.p.parse_term1("*"), "*")
        self.assertEqual(self.p.parse_term3("Ter"), "Ter")
        self.assertEqual(self.p.parse_term13("*"), "*")
        self.assertEqual(self.p.parse_term13("Ter"), "Ter")

        # AAs
        self.assertEqual(self.p.parse_aa1("A"), "A")
        self.assertEqual(self.p.parse_aa3("Ala"), "Ala")
        self.assertEqual(self.p.parse_aa13("A"), "A")
        self.assertEqual(self.p.parse_aa13("Ala"), "Ala")

        # mixto
        self.assertEqual(self.p.parse_aat1("*"), "*")
        self.assertEqual(self.p.parse_aat1("A"), "A")
        self.assertEqual(self.p.parse_aat3("Ter"), "Ter")
        self.assertEqual(self.p.parse_aat3("Ala"), "Ala")
        self.assertEqual(self.p.parse_aat13("*"), "*")
        self.assertEqual(self.p.parse_aat13("A"), "A")
        self.assertEqual(self.p.parse_aat13("Ter"), "Ter")
        self.assertEqual(self.p.parse_aat13("Ala"), "Ala")

    def test_parser_aa_seq(self):
        # mixto
        self.assertEqual(self.p.parse_aat1_seq("*"), "*")
        self.assertEqual(self.p.parse_aat1_seq("A"), "A")
        self.assertEqual(self.p.parse_aat3_seq("Ter"), "Ter")
        self.assertEqual(self.p.parse_aat3_seq("Ala"), "Ala")
        self.assertEqual(self.p.parse_aat13_seq("*"), "*")
        self.assertEqual(self.p.parse_aat13_seq("A"), "A")
        self.assertEqual(self.p.parse_aat13_seq("Ter"), "Ter")
        self.assertEqual(self.p.parse_aat13_seq("Ala"), "Ala")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pprint
import re
import sys
import os

import unittest

import pytest

from hgvs.exceptions import HGVSError, HGVSDataNotAvailableError, HGVSParseError, HGVSInvalidVariantError
from hgvs.enums import Datum
import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.normalizer
import hgvs.parser
import hgvs.sequencevariant
import hgvs.validator
import hgvs.variantmapper
from support import CACHE


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    def setUp(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=False)
        self.vm_rr = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=True)
        self.hp = hgvs.parser.Parser()
        self.hn = hgvs.normalizer.Normalizer(self.hdp)
        self.hv = hgvs.validator.IntrinsicValidator()
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh37', alt_aln_method='splign')
        self.am38 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh38', alt_aln_method='splign')
        self.vn = hgvs.normalizer.Normalizer(self.hdp, shuffle_direction=3, cross_boundaries=True)

    def test_424_430_nochange_parse_and_format(self):
        h = "NM_012.3:c.1="
        v = self.hp.parse_hgvs_variant(h)
        self.assertEqual("NM_012.3:c.1=", str(v))
        self.assertEqual("", v.posedit.edit.ref)
        self.assertEqual("", v.posedit.edit.alt)

        h = "NM_012.3:c.1A="
        v = self.hp.parse_hgvs_variant(h)
        self.assertEqual("NM_012.3:c.1=", str(v))
        self.assertEqual("A", v.posedit.edit.ref)
        self.assertEqual("A", v.posedit.edit.alt)

    def test_459_exception_when_ac_nonexistent(self):
        bogus_ac = "NM_000000.99"
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.seqfetcher.fetch_seq(bogus_ac)

    def test_499_whole_gene_dup(self):
        # Verify that 1_*1dup works
        self.am37.c_to_p(self.hp.parse_hgvs_variant("NM_001637.3:c.1_*1dup"))

        # Now try -1_*1dup (essentially, this is issue #499)
        self.am37.c_to_p(self.hp.parse_hgvs_variant("NM_001637.3:c.-1_*1dup"))



#!/usr/bin/env python

import hgvs
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.sequencevariant
import hgvs.assemblymapper

hdp = hgvs.dataproviders.uta.connect()
hp = hgvs.parser.Parser()
evm = hgvs.assemblymapper.AssemblyMapper(hdp,
                                           replace_reference=True, assembly_name='GRCh37',
                                           alt_aln_method='splign')

#v = hp.parse_hgvs_variant("NM_000059.3:c.7790A>G")
#v = hp.parse_hgvs_variant("NM_000059.3:c.7790_7792delAAG")
v = hp.parse_hgvs_variant("NM_000059.3:c.7790delAAG")
evm.c_to_p(v)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

import pytest

import hgvs.hgvsposition


class Test_HGVSPosition(unittest.TestCase):
    @pytest.mark.quick
    @pytest.mark.models
    def test_hgvsposition(self):
        var = hgvs.hgvsposition.HGVSPosition(
            ac="NM_01234.5",
            type="c",
            pos=hgvs.location.Interval(
                hgvs.location.BaseOffsetPosition(base=12, offset=+34),
                hgvs.location.BaseOffsetPosition(base=56, offset=-78)))

        self.assertEqual(str(var), "NM_01234.5:c.12+34_56-78")


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pprint
import re
import sys
import os

import unittest

import pytest

from hgvs.exceptions import HGVSError, HGVSDataNotAvailableError, HGVSParseError, HGVSInvalidVariantError
from hgvs.enums import Datum
import hgvs.assemblymapper
import hgvs.dataproviders.uta
import hgvs.normalizer
import hgvs.parser
import hgvs.sequencevariant
import hgvs.validator
import hgvs.variantmapper
from support import CACHE


@pytest.mark.issues
class Test_Issues(unittest.TestCase):
    def setUp(self):
        self.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "run"), cache=CACHE)
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=False)
        self.vm_rr = hgvs.variantmapper.VariantMapper(self.hdp, replace_reference=True)
        self.hp = hgvs.parser.Parser()
        self.hn = hgvs.normalizer.Normalizer(self.hdp)
        self.hv = hgvs.validator.IntrinsicValidator()
        self.am37 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh37', alt_aln_method='splign')
        self.am38 = hgvs.assemblymapper.AssemblyMapper(
            self.hdp, replace_reference=True, assembly_name='GRCh38', alt_aln_method='splign')
        self.vn = hgvs.normalizer.Normalizer(self.hdp, shuffle_direction=3, cross_boundaries=True)

    def test_525(self):
        """https://github.com/biocommons/hgvs/issues/525"""

        # simple test case
        hgvs = "NM_001637.3:c.3_4insTAG"    # insert stop in phase at AA 2
        var_c = self.hp.parse_hgvs_variant(hgvs)
        var_p = self.am38.c_to_p(var_c)
        assert str(var_p) == "NP_001628.1:p.(Gln2Ter)"



r"""Test support for optional strict bounds on transcript coordinates

https://github.com/biocommons/hgvs/issues/437

    RMPR (- strand, chr 9)

    NC_000009.12          7751    7759  //      8018           8033      8043
    NC_000009.11          7748    7756  //      8015           8030      8040
                             |       |  //         |              |         |
               +      acaaaaaACAGCCGCG  //  CACGAACCacgtcctcagcttcacagagtagtatt
               -      tgtttttTGTCGGCGC  //  GTGCTTGGtgcaggagtcgaagtgtctcatcataa
                      aaaaaaaTGTCGGCGC  //  GTGCTTGG                 
                             |       |  //         |\        |         |
    NR_003051.3            268     260  //        <1 -1    -10       -20

    tests:                  ^^                     ^^                >-------<
                            43                     21                   dup

"""

import unittest

import pytest

import hgvs
import hgvs.alignmentmapper
import hgvs.variantmapper
from hgvs.exceptions import HGVSInvalidVariantError




@pytest.mark.usefixtures("kitchen_sink_setup")
class Test437_RMRP(unittest.TestCase):

    def setUp(self):
        # Generate n. and g. variants at terminal positions of tx, and
        # variants that are 1 base out of bounds

        # "NR_003051.3:n.1G>T", "NC_000009.11:g.35658015C>A"
        # "NR_003051.3:n.268T>C", "NC_000009.11:g.35657748A>G"

        # "NR_003051.3:n.-1G>C", "NC_000009.11:g.35658016A>G"
        # "NR_003051.3:n.269N>C", "NC_000009.11:g.35657747A>G"
        # "NR_003051.3:n.-19_-18insACT", "NC_000009.11:g.35658038_35658040dup"   # ~= 40_41insCAT; NB rotated 1


        self.s1_n = self.parser.parse("NR_003051.3:n.1G>T")
        self.e1_n = self.parser.parse("NR_003051.3:n.268T>C")
        
        self.s2_n = self.parser.parse("NR_003051.3:n.-1G>C")
        self.e2_n = self.parser.parse("NR_003051.3:n.269N>C")
        
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp)
        self.alm37 = hgvs.alignmentmapper.AlignmentMapper(
            self.hdp, "NR_003051.3", "NC_000009.11", "splign")
        self.alm38 = hgvs.alignmentmapper.AlignmentMapper(
            self.hdp, "NR_003051.3", "NC_000009.12", "splign")



    def test_tx_start(self):
        hgvs_n = "NR_003051.3:n.1G>T"
        hgvs_g = "NC_000009.11:g.35658015C>A"

        var_n = self.parser.parse(hgvs_n)
        var_g = self.parser.parse(hgvs_g)

        var_ng = self.am37.n_to_g(var_n)
        assert var_ng.posedit.pos == var_g.posedit.pos
        assert str(var_ng) == hgvs_g
    
        var_gn = self.am37.g_to_n(var_g, var_n.ac)
        assert var_gn.posedit.pos == var_n.posedit.pos
        assert str(var_gn) == hgvs_n



    def test_terminii(self):
        hgvs.global_config.mapping.strict_bounds = True

        # Sanity check projections to genome for in-bound variations
        # n -> g -> n roundtrip
        s1_g = self.am37.n_to_g(self.s1_n)
        assert str(s1_g) == "NC_000009.11:g.35658015C>A"
        assert self.s1_n == self.am37.g_to_n(s1_g, self.s1_n.ac)

        e1_g = self.am37.n_to_g(self.e1_n)
        assert str(e1_g) == "NC_000009.11:g.35657748A>G"
        assert self.e1_n == self.am37.g_to_n(e1_g, self.e1_n.ac)


        # Disable bounds checking and retry
        hgvs.global_config.mapping.strict_bounds = False

        s2_g = self.am37.n_to_g(self.s2_n)
        assert str(s2_g) == "NC_000009.11:g.35658016A>G"
        assert s2_g.posedit.pos.start.base - s1_g.posedit.pos.start.base == 1
        #BUG: assert self.s2_n == self.am37.g_to_n(s2_g, self.s2_n.ac)

        e2_g = self.am37.n_to_g(self.e2_n)
        assert str(e2_g) == "NC_000009.11:g.35657747A>G"
        assert e2_g.posedit.pos.start.base - e1_g.posedit.pos.start.base == -1
        #BUG: assert self.e2_n == self.am37.g_to_n(e2_g, self.e2_n.ac)

        hgvs.global_config.mapping.strict_bounds = True


    def test_enforce_strict_bounds(self):
        """Ensure that an exception is raised when outside bounds"""

        # TODO: These exceptions are raised, but with the wrong message
        # use `match=` arg
        with pytest.raises(HGVSInvalidVariantError):
            s2_g = self.am37.n_to_g(self.s2_n)

        with pytest.raises(HGVSInvalidVariantError):
            e2_g = self.am37.n_to_g(self.e2_n)



def test_oob_dup(parser, am37):
    """Intentionally preserve dup, derived from genomic sequence, when
    projecting to out-of-bounds transcript coordinates

    """
    hgvs.global_config.mapping.strict_bounds = False

    # n_to_g
    var_n = parser.parse("NR_003051.3:n.-19_-18insACT")
    var_g = am37.n_to_g(var_n)
    assert str(var_g) == "NC_000009.11:g.35658038_35658040dup"

    # g_to_n
    var_n2 = am37.g_to_n(var_g, "NR_003051.3")
    assert str(var_n2) == "NR_003051.3:n.-21_-19dup"

    hgvs.global_config.mapping.strict_bounds = True



if __name__ == "__main__":
    from hgvs.easy import *
    test_437(parser=parser, am=am37)



"""translate between HGVS and other formats"""

import bioutils.assemblies

import hgvs.normalizer


def _as_interbase(posedit):
    if posedit.edit.type == "ins":
        # ins coordinates (only) exclude left position
        start_i = posedit.pos.start.base
        end_i = posedit.pos.end.base - 1
    else:
        start_i = posedit.pos.start.base - 1
        end_i = posedit.pos.end.base
    return (start_i, end_i)


class Babelfish:
    def __init__(self, hdp, assembly_name):
        self.hdp = hdp
        self.hn = hgvs.normalizer.Normalizer(hdp,
                                             cross_boundaries=False,
                                             shuffle_direction=5,
                                             validate=False)
        self.ac_to_chr_name_map = {
            sr["refseq_ac"]: sr["name"]
            for sr in bioutils.assemblies.get_assembly("GRCh38")["sequences"]}



    def hgvs_to_vcf(self, var_g):
        """**EXPERIMENTAL**

        converts a single hgvs allele to (chr, pos, ref, alt) using
        the given assembly_name. The chr name uses official chromosome
        name (i.e., without a "chr" prefix).

        Returns None for non-variation (e.g., NC_000006.12:g.49949407=)

        """

        if var_g.type != "g":
            raise RuntimeError("Expected g. variant, got {var_g}".format(var_g=var_g))

        vleft = self.hn.normalize(var_g)

        (start_i, end_i) = _as_interbase(vleft.posedit)

        chr = self.ac_to_chr_name_map[vleft.ac]

        typ = vleft.posedit.edit.type

        if typ == "dup":
            start_i -= 1
            alt = self.hdp.seqfetcher.fetch_seq(vleft.ac, start_i, end_i)
            ref = alt[0]
            end_i = start_i
            return (chr, start_i + 1, ref, alt, typ)

        if vleft.posedit.edit.ref == vleft.posedit.edit.alt:
            return None

        alt = vleft.posedit.edit.alt or ""

        if end_i - start_i == 1 and vleft.posedit.length_change == 0:
            # SNVs aren't left anchored
            ref = vleft.posedit.edit.ref

        else:
            # everything else is left-anchored
            start_i -= 1
            ref = self.hdp.seqfetcher.fetch_seq(vleft.ac, start_i, end_i)
            alt = ref[0] + alt

        return (chr, start_i + 1, ref, alt, typ)



if __name__ == "__main__":
    """
      49949___  400       410       420
                  |123456789|123456789|
    NC_000006.12  GACCAGAAAGAAAAATAAAAC

    """

    import hgvs.easy
    import hgvs.normalizer
    from hgvs.extras.babelfish import Babelfish

    babelfish38 = Babelfish(hgvs.easy.hdp, assembly_name="GRCh38")
    hnl = hgvs.normalizer.Normalizer(hgvs.easy.hdp,
                                     cross_boundaries=False, shuffle_direction=5, validate=False)

    def _h2v (h):
        return babelfish38.hgvs_to_vcf(hgvs.easy.parser.parse(h))
    
    def _vp(h):
        v = hgvs.easy.parser.parse(h)
        vl = hnl.normalize(v)
        return (v,vl)

    for h in (
            # Non-variation
            "NC_000006.12:g.49949407=",
            
            # SNV
            "NC_000006.12:g.49949407A>T",

            # delins
            "NC_000006.12:g.49949413_49949414delinsCC",

            # del
            "NC_000006.12:g.49949415del",
            "NC_000006.12:g.49949413del",
            "NC_000006.12:g.49949414del",
            "NC_000006.12:g.49949413_49949414del",

            # ins
            "NC_000006.12:g.49949413_49949414insC",
            "NC_000006.12:g.49949414_49949415insC",
            "NC_000006.12:g.49949414_49949415insCC",

            # ins (dup)
            "NC_000006.12:g.49949413_49949414insA",
            "NC_000006.12:g.49949414_49949415insA",
            "NC_000006.12:g.49949414_49949415insAA"
            ):
        print('assert _h2v("{h}") == {res}'.format(res=str(_h2v(h)), h=h))
